import wx
import wx.lib.scrolledpanel as scrolledPanel
import wx.lib.dialogs as dialogs
import os
import re
import sys

sys.path.append(os.path.dirname(os.path.realpath(__file__)))

from util.dataHandling import DataHandling
from util.metaDataSet import MetaDataSet
from util.utils import Cardinality, Datatype, Validity
from util.metaDataHelpers import CalendarDlg


def collectMetadata():
    """
    Runner function that launches the app.

    Calling this method initiates a data handler and opens the GUI.

    Example:
        To run the application, simply call:
        ```
        $ collectMetadata()
        ```
    """
    # create a data handler
    global data_handler
    data_handler = DataHandling()
    # open GUI
    app = wx.App()
    ProjectFrame()
    app.MainLoop()


class ProjectFrame(wx.Frame):
    def __init__(self):
        """
        This class holds the frame of the main application window.
        """
        super().__init__(parent=None,
                         title='Project Data Editor', size=(1100, 750))
        self.panel = ProjectPanel(self)
        self.__create_menu()
        self.Show()

    def __create_menu(self):
        """
        Create the menu, add a folder dialog box
        """
        menu_bar = wx.MenuBar()
        file_menu = wx.Menu()
        open_folder_menu_item = file_menu.Append(wx.ID_NEW, 'Add new Project', 'Open a folder with project files')
        self.Bind(event=wx.EVT_MENU, handler=self.panel.on_add_new_project, source=open_folder_menu_item)
        save_menu_item = file_menu.Append(wx.ID_SAVE, "&Save")
        self.Bind(wx.EVT_MENU, self.__on_save, source=save_menu_item)
        menu_bar.Append(file_menu, '&File')
        options_menu = wx.Menu()
        menu_bar.Append(options_menu, '&Options')
        options_help = wx.Menu()
        menu_bar.Append(options_help, '&Help')
        self.SetMenuBar(menu_bar)

    def __on_save(self, event):
        """
        Menu item for saving the current window
        """
        if data_handler.current_window:
            data_handler.current_window.save()
        else:
            data_handler.save_data()


class ProjectPanel(wx.Panel):
    def __init__(self, parent: ProjectFrame):
        """
        Panel containing the contents of the application's main window.

        The Panel displays a list of projects on top.
        On the bottom left it shows a preview of the turtle serialization of the selected project.
        On the bottom right it holds a number of buttons for various tasks.
        (Adding/removing projects, modifying projects, exporting data, etc.)

        Args:
            parent (ProjectFrame): The parent frame to hold this panel.
        """
        super().__init__(parent)
        self.folder_path = ""
        self.row_obj_dict = {}
        self.project_dependent_buttons = []

        # Create list for projects
        main_sizer = wx.FlexGridSizer(3, 1, 10, 10)
        # main_sizer = wx.BoxSizer(wx.VERTICAL)
        title = wx.StaticText(self, label="DaSCH Service Platform - Metadata Collection")
        main_sizer.Add(title, 0, wx.ALL | wx.LEFT, 10)
        self.list_ctrl = wx.ListCtrl(self, size=(-1, 200), style=wx.LC_REPORT | wx.BORDER_SUNKEN)
        self.list_ctrl.Bind(wx.EVT_LIST_ITEM_SELECTED, self.on_item_selected)
        self.list_ctrl.Bind(wx.EVT_LIST_ITEM_DESELECTED, self.on_item_selected)
        main_sizer.Add(self.list_ctrl, 0, wx.ALL | wx.EXPAND, 10)

        # Create turtle preview
        bottom_sizer = wx.FlexGridSizer(1, 2, 10, 10)
        bottom_sizer.AddGrowableCol(0)
        bottom_sizer.AddGrowableRow(0)
        rdf_display = wx.TextCtrl(self, value="No Project selected.",
                                  style=wx.TE_READONLY | wx.TE_MULTILINE, size=(400, -1))
        self.rdf_display = rdf_display
        bottom_sizer.Add(rdf_display, flag=wx.EXPAND)

        # Create Buttons
        button_sizer = wx.BoxSizer(wx.VERTICAL)
        new_folder_button = wx.Button(self, label='Add new Project')
        new_folder_button.Bind(wx.EVT_BUTTON, self.on_add_new_project)
        button_sizer.Add(new_folder_button, 0, wx.ALL | wx.EXPAND, 7)

        import_project_button = wx.Button(self, label='Import Project')
        import_project_button.Bind(wx.EVT_BUTTON, self.on_import_project)
        button_sizer.Add(import_project_button, 0, wx.ALL | wx.EXPAND, 7)

        remove_folder_button = wx.Button(self, label='Remove selected Project')
        self.project_dependent_buttons.append(remove_folder_button)
        remove_folder_button.Bind(wx.EVT_BUTTON, self.on_remove_project)
        button_sizer.Add(remove_folder_button, 0, wx.ALL | wx.EXPAND, 7)

        edit_tabs_button = wx.Button(self, label='Edit selected Project')
        self.project_dependent_buttons.append(edit_tabs_button)
        edit_tabs_button.Bind(wx.EVT_BUTTON, self.on_edit_tabbed)
        button_sizer.Add(edit_tabs_button, 0, wx.ALL | wx.EXPAND, 7)

        validate_button = wx.Button(self, label='Validate selected Project')
        self.project_dependent_buttons.append(validate_button)
        validate_button.Bind(wx.EVT_BUTTON, self.on_validate)
        button_sizer.Add(validate_button, 0, wx.ALL | wx.EXPAND, 7)

        process_xml_button = wx.Button(self, label='Export selected Project as RDF')
        self.project_dependent_buttons.append(process_xml_button)
        process_xml_button.Bind(wx.EVT_BUTTON, self.export_data)
        button_sizer.Add(process_xml_button, 0, wx.ALL | wx.EXPAND, 7)

        zip_and_export_btn = wx.Button(self, label='ZIP and Export Project')
        self.project_dependent_buttons.append(zip_and_export_btn)
        zip_and_export_btn.Bind(wx.EVT_BUTTON, self.__on_zip_and_export)
        button_sizer.Add(zip_and_export_btn, 0, wx.ALL | wx.EXPAND, 7)

        # Layout it all
        bottom_sizer.Add(button_sizer, 0, wx.ALL, 10)
        main_sizer.Add(bottom_sizer, 0, wx.ALL | wx.EXPAND, 10)
        main_sizer.AddGrowableRow(2)
        self.SetSizer(main_sizer)
        self.Fit()
        self.create_header()
        self.refresh_view()
        self.Layout()

    def __on_zip_and_export(self, event):
        """
        Export a ZIP archive containing the selected files, metadata and a pickle with the binaries.
        """
        title = "Where to export to?"
        with wx.DirDialog(self, title, style=wx.DD_DEFAULT_STYLE) as dlg:
            if dlg.ShowModal() == wx.ID_OK:
                path = dlg.GetPath()
                print(path)
                data_handler.zip_and_export(self.get_selected_project(), path)

    def on_add_new_project(self, event):
        """
        Open a new folder and add it to projects.
        """
        title = "Enter Project Shortcode:\n(4 alphanumeric characters)\n\nIf your project doesn't have a shortcode assigned yet, \nplease contact the DaSCH Client Services)"
        dlg = wx.TextEntryDialog(self, message=title, caption="Enter Shortcode")
        if dlg.ShowModal() == wx.ID_OK:
            shortcode = dlg.GetValue()
            if not re.match('[a-zA-Z0-9]{4}$', shortcode):
                print("Invalid shortcode entered.")
                return
        else:
            return
        title = "Choose a directory:"
        dlg = wx.DirDialog(self, title, style=wx.DD_DEFAULT_STYLE)
        if dlg.ShowModal() == wx.ID_OK:
            self.add_new_project(dlg.GetPath(), shortcode)
        dlg.Destroy()

    def refresh_repos(self):
        """
        Refresh all loaded repos in the list.
        """
        for i, project in enumerate(data_handler.projects):
            if i < self.list_ctrl.GetItemCount():
                self.list_ctrl.SetItem(i, 0, project.path)
            else:
                self.list_ctrl.InsertItem(i, project.path)
            self.list_ctrl.SetItem(i, 1, project.name)
            self.list_ctrl.SetItem(i, 2, project.shortcode)
            self.list_ctrl.SetItem(i, 3, str(project.files))
            self.list_ctrl.SetItem(i, 4, project.get_status())

    def create_header(self):
        """
        Here we create the header
        """
        # Construct a header
        self.list_ctrl.InsertColumn(0, 'Folder', width=300)
        self.list_ctrl.InsertColumn(1, 'Project', width=180)
        self.list_ctrl.InsertColumn(2, 'Shortcode', width=90)
        self.list_ctrl.InsertColumn(3, 'List of files', width=340)
        self.list_ctrl.InsertColumn(4, 'Status', width=350)

    def refresh_view(self):
        self.refresh_repos()
        self.display_rdf()
        self.refresh_buttons()
        self.Layout()

    def refresh_buttons(self):
        if self.get_selected_project():
            flag = True
        else:
            flag = False
        for b in self.project_dependent_buttons:
            if flag:
                b.Enable()
            else:
                b.Disable()

    def display_rdf(self):
        project = self.get_selected_project()
        if project:
            txt = project.get_turtle()
        else:
            txt = "No project selected."
        self.rdf_display.SetValue(txt)

    def get_selected_project(self) -> MetaDataSet:
        selection = self.list_ctrl.GetFirstSelected()
        if selection < 0:
            return
        shortcode = self.list_ctrl.GetItem(selection, col=2).GetText()
        return data_handler.get_project_by_shortcode(shortcode)

    def on_edit_tabbed(self, event):
        """
        This function calls the EditBaseDialog and hands over pFiles, a list.
        """
        repo = self.get_selected_project()
        if repo:
            window = TabbedWindow(self, repo)
            data_handler.current_window = window
            window.Show()
            self.Disable()

    def export_data(self, event):
        """ Set selection and call create_xml """
        selection = self.list_ctrl.GetFocusedItem()
        if selection >= 0:
            res = data_handler.validate_and_export_data(selection)
            print(res[0])
            # LATER: let this return indication of success. display something to the user.

    def add_new_project(self, folder_path, shortcode):
        """ Add a new project.

            Where is this function called? It is called by on_add_new_project in in the Class ProjectFrame
            What should this function do? It should get a new project, store it and then reload the project list
        """
        dir_list = os.listdir(folder_path)
        if '.DS_Store' in dir_list:
            dir_list.remove('.DS_Store')
        data_handler.add_project(folder_path, shortcode, dir_list)
        self.refresh_view()

    def on_import_project(self, event):
        with wx.FileDialog(self, "Choose file:",
                           style=wx.FD_OPEN | wx.FD_FILE_MUST_EXIST) as fd:
            if fd.ShowModal() == wx.ID_OK:
                f = fd.GetPath()
                if not f.endswith('.data'):
                    print(f'Could not import file: {f}\n  For now, only .data supported.')
                    # LATER: allow RDF import here
                    return
                data_handler.import_project(f)
                self.refresh_view()

    def on_remove_project(self, event):
        selected = self.get_selected_project()
        msg = f"Are you sure you want to delete the Project '{selected.name} ({selected.shortcode})'"
        with wx.MessageDialog(self, "Sure?", msg, wx.YES_NO) as dlg:
            if dlg.ShowModal() == wx.ID_YES:
                data_handler.remove_project(selected)
                self.list_ctrl.DeleteAllItems()
                self.refresh_view()

    def on_validate(self, event):
        repo = self.get_selected_project()
        if repo:
            conforms, results_graph, results_text = data_handler.validate_graph(repo)
            if conforms:
                with wx.MessageDialog(self, "Validation Successful", "Validation Successful", wx.OK | wx.ICON_INFORMATION) as dlg:
                    dlg.ShowModal()
            else:
                with dialogs.ScrolledMessageDialog(self, results_text, "Validation Failed", size=(800, 500)) as dlg:
                    dlg.ShowModal()

    def on_item_selected(self, event):
        # LATER: look into why this is called twice, which makes it slow
        # print('item selected')
        self.refresh_view()


class TabOne(wx.Panel):
    """
    Tab holding the project base information
    """

    def __init__(self, parent, dataset):
        wx.Panel.__init__(self, parent)
        self.dataset = dataset

        # Project name as caption
        sizer = wx.GridBagSizer(10, 10)
        project_label = wx.StaticText(self, label="Current Project:")
        project_name = wx.StaticText(self, label=self.dataset.name)
        sizer.Add(project_label, pos=(0, 0))
        sizer.Add(project_name, pos=(0, 1))

        # Path to folder
        path_label = wx.StaticText(self, label="Path: ")
        sizer.Add(path_label, pos=(1, 0))
        path_field = wx.TextCtrl(self, style=wx.TE_READONLY, size=(550, -1))
        path_field.SetValue(self.dataset.path)
        sizer.Add(path_field, pos=(1, 1))
        path_help = wx.Button(self, label="?")
        path_help.Bind(wx.EVT_BUTTON, lambda event: self.show_help(event,
                                                                   "Path to the folder with the data",
                                                                   "/some/path/to/folder"))
        sizer.Add(path_help, pos=(1, 2))

        # Files
        files_label = wx.StaticText(self, label="Files: ")
        sizer.Add(files_label, pos=(2, 0))
        data_sizer = wx.BoxSizer()
        file_list = wx.ListBox(self, size=(550, -1))
        for f in dataset.files:
            file_list.Append(f)
        data_sizer.Add(file_list)
        button_sizer = wx.BoxSizer(wx.VERTICAL)
        btn_add = wx.Button(self, label="Add File(s)")
        btn_add.Bind(wx.EVT_BUTTON, lambda event: self.add_file(
            dataset, file_list))
        btn_del = wx.Button(self, label="Remove Selected")
        btn_del.Bind(wx.EVT_BUTTON, lambda event: self.remove_file(
            dataset, file_list))
        button_sizer.Add(btn_add, flag=wx.EXPAND)
        button_sizer.Add(btn_del, flag=wx.EXPAND)
        data_sizer.Add(button_sizer)
        sizer.Add(data_sizer, pos=(2, 1))
        path_help = wx.Button(self, label="?")
        path_help.Bind(wx.EVT_BUTTON, lambda event: self.show_help(event,
                                                                   "Files associated with the project",
                                                                   "sample_project.zip"))
        sizer.Add(path_help, pos=(2, 2))
        sizer.AddGrowableCol(1)
        self.SetSizer(sizer)

    def show_help(self, evt, message, sample):
        msg = f"Description:\n{message}\n\nExample:\n{sample}"
        win = HelpPopup(self, msg)
        btn = evt.GetEventObject()
        pos = btn.ClientToScreen((0, 0))
        sz = btn.GetSize()
        win.Position(pos, (0, sz[1]))
        win.Popup()

    def add_file(self, dataset, listbox):
        with wx.FileDialog(self, "Choose file(s):",
                           style=wx.FD_OPEN | wx.FD_FILE_MUST_EXIST | wx.FD_MULTIPLE) as fd:
            if fd.ShowModal() == wx.ID_OK:
                for p in fd.GetPaths():
                    f = str(p)
                    if f.startswith(dataset.path):
                        f = f.replace(f'{dataset.path}/', '')
                    else:
                        continue
                    if f and f not in dataset.files:
                        dataset.files.append(f)
                        listbox.Append(f)

    def remove_file(self, dataset, file_list):
        selection = file_list.GetSelection()
        if selection >= 0:
            string_selected = file_list.GetString(selection)
            dataset.files.remove(string_selected)
            file_list.Delete(selection)


class PropertyRow():

    def __init__(self, parent, prop, sizer, index, metadataset):
        """
        A row in a tab of the UI

        This Class organizes a single row in the data tabs.
        Upon initiation, the UI elements ara generated and placed.
        Later on, the data handler can let this class return the value that the property should be assigned.

        Args:
            parent (wx.ScrolledWindow): The scrolled panel in which the row is to be placed.
            data_class (Project|Dataset|List[Person]|etc.): The Class that is to be displayed
            prop (Property): The property to be displayed
            sizer (wx.Sizer): The sizer that organizes the layout of the parent component
            index (int): the row in the sizer grid
        """
        self.prop_name = prop.name
        self.metadataset = metadataset
        self.data_widget = None
        self.choice_widget = None
        self.parent = parent
        self.validity_msg = ""
        name_label = wx.StaticText(parent, label=prop.name + ": ")
        sizer.Add(name_label)
        if prop.datatype == Datatype.STRING \
                or prop.datatype == Datatype.STRING_OR_URL \
                or prop.datatype == Datatype.URL \
                or prop.datatype == Datatype.EMAIL \
                or prop.datatype == Datatype.DOWNLOAD \
                or prop.datatype == Datatype.PLACE:
            self.__setup_string(parent, sizer, prop)
        elif prop.datatype == Datatype.DATE:
            if prop.cardinality == Cardinality.ONE \
                    or prop.cardinality == Cardinality.ZERO_OR_ONE:
                inner_sizer = wx.BoxSizer()
                date = wx.StaticText(parent, size=(100, -1))
                pick_date_button = wx.Button(parent, label="Pick Date")
                pick_date_button.Bind(
                    wx.EVT_BUTTON, lambda event: parent.pick_date(event, date, self.prop))
                inner_sizer.Add(date)
                inner_sizer.Add(pick_date_button)
                sizer.Add(inner_sizer, flag=wx.EXPAND)
                self.data_widget = date
        elif prop.datatype == Datatype.PROJECT:
            txt = wx.StaticText(parent, label=str(prop.value))
            self.data_widget = txt
            sizer.Add(txt, flag=wx.EXPAND)
        elif prop.datatype == Datatype.PERSON_OR_ORGANIZATION or \
                prop.datatype == Datatype.PERSON or \
                prop.datatype == Datatype.ORGANIZATION or \
                prop.datatype == Datatype.GRANT or \
                prop.datatype == Datatype.CONTROLLED_VOCABULARY:
            self.__setup_dropdown(parent, sizer, prop)
        elif prop.datatype == Datatype.DATA_MANAGEMENT_PLAN:
            inner_sizer = wx.BoxSizer(wx.VERTICAL)
            cb = wx.CheckBox(parent, label='is available')
            cb.Bind(wx.EVT_CHECKBOX, lambda e: data_handler.update_all())
            inner_sizer.Add(cb, flag=wx.EXPAND)
            text = wx.TextCtrl(parent, style=wx.TE_PROCESS_ENTER)
            text.SetHint('Optional URL')
            text.Bind(wx.EVT_TEXT_ENTER, self.onValueChange)
            inner_sizer.Add(text, flag=wx.EXPAND)
            sizer.Add(inner_sizer, flag=wx.EXPAND)
            self.data_widget = [cb, text]
        elif prop.datatype == Datatype.ADDRESS:
            self.__setup_address(parent, sizer)
        elif prop.datatype == Datatype.ATTRIBUTION:
            self.__setup_attribution(parent, sizer)
        elif prop.datatype == Datatype.SHORTCODE:
            text = wx.StaticText(parent)
            self.data_widget = text
            sizer.Add(text, flag=wx.EXPAND)

        btn = wx.Button(parent, label="?")
        btn.Bind(wx.EVT_BUTTON, lambda event: parent.show_help(event, prop.description, prop.example))
        sizer.Add(btn)
        opt = wx.Button(parent, label=Cardinality.get_optionality_string(prop.cardinality))
        opt.Bind(wx.EVT_BUTTON, lambda event: parent.show_validity(event,
                                                                   self.validity_msg,
                                                                   Cardinality.as_sting(prop.cardinality)))
        self.validity_widget = opt
        sizer.Add(opt, flag=wx.RIGHT, border=5)
        self.refresh_ui()

    def __setup_string(self, parent, sizer, prop):
        if prop.cardinality == Cardinality.ONE \
                or prop.cardinality == Cardinality.ZERO_OR_ONE:  # String or similar, exactly 1 or 0-1
            self.__setup_string_one_card(parent, sizer, prop)
        elif prop.cardinality == Cardinality.ONE_TO_TWO \
                or prop.cardinality == Cardinality.ZERO_TO_TWO:
            self.__setup_string_two_card(parent, sizer, prop)
        elif prop.cardinality == Cardinality.ONE_TO_UNBOUND \
                or prop.cardinality == Cardinality.ONE_TO_UNBOUND_ORDERED \
                or prop.cardinality == Cardinality.UNBOUND:  # String or similar, 1-n or 0-n
            self.__setup_string_multi_card(parent, sizer, prop)

    def __setup_string_multi_card(self, parent, sizer, prop):
        scroller = wx.lib.scrolledpanel.ScrolledPanel(parent)
        scroll_sizer = wx.BoxSizer(wx.VERTICAL)
        if prop.multiline:
            inner_sizer = wx.BoxSizer()
            text_sizer = wx.BoxSizer(wx.VERTICAL)
            textcontrol = wx.TextCtrl(scroller, style=wx.TE_PROCESS_ENTER)
            textcontrol.Bind(wx.EVT_TEXT_ENTER,
                             lambda e: parent.add_to_list(e,
                                                          content_list,
                                                          textcontrol,
                                                          textcontrol.GetValue()))
            text_sizer.Add(textcontrol, flag=wx.EXPAND)
            text_sizer.AddSpacer(5)
            content_list = wx.ListBox(scroller)
            text_sizer.Add(content_list, 1, flag=wx.EXPAND)
            inner_sizer.Add(text_sizer, 1, flag=wx.EXPAND)
            inner_sizer.AddSpacer(5)
            button_sizer = wx.BoxSizer(wx.VERTICAL)
            plus_button = wx.Button(scroller, label="+")
            plus_button.Bind(wx.EVT_BUTTON,
                             lambda e: parent.add_to_list(e,
                                                          content_list,
                                                          textcontrol,
                                                          textcontrol.GetValue()))
            button_sizer.Add(plus_button, flag=wx.EXPAND)

            remove_button = wx.Button(scroller, label="Del Selected")
            remove_button.Bind(wx.EVT_BUTTON,
                               lambda event: parent.remove_from_list(event,
                                                                     content_list))
            button_sizer.Add(remove_button)
            inner_sizer.Add(button_sizer)
        else:
            inner_sizer = wx.BoxSizer()
            textcontrol = wx.TextCtrl(scroller, style=wx.TE_PROCESS_ENTER, size=(300, -1))
            textcontrol.Bind(wx.EVT_TEXT_ENTER,
                             lambda e: parent.add_to_list(e,
                                                          content_list,
                                                          textcontrol,
                                                          textcontrol.GetValue()))
            inner_sizer.Add(textcontrol, 0)
            inner_sizer.AddSpacer(5)
            button_sizer = wx.BoxSizer(wx.VERTICAL)
            plus_button = wx.Button(scroller, label="+")
            plus_button.Bind(wx.EVT_BUTTON,
                             lambda e: parent.add_to_list(e,
                                                          content_list,
                                                          textcontrol,
                                                          textcontrol.GetValue()))
            button_sizer.Add(plus_button, flag=wx.EXPAND)

            remove_button = wx.Button(scroller, label="Del Selected")
            remove_button.Bind(wx.EVT_BUTTON,
                               lambda event: parent.remove_from_list(event,
                                                                     content_list))
            button_sizer.Add(remove_button)
            inner_sizer.Add(button_sizer)
            inner_sizer.AddSpacer(5)
            content_list = wx.ListBox(scroller)
            inner_sizer.Add(content_list, 1, flag=wx.EXPAND)
        # sizer.Add(inner_sizer, flag=wx.EXPAND)
        scroll_sizer.Add(inner_sizer, flag=wx.EXPAND)
        scroller.Sizer = scroll_sizer
        scroller.SetupScrolling(scroll_x=True, scroll_y=False)
        sizer.Add(scroller, flag=wx.EXPAND)
        self.data_widget = content_list

    def __setup_string_two_card(self, parent, sizer, prop):
        if prop.cardinality == Cardinality.ONE_TO_TWO:  # String or similar, 1-2
            inner_sizer = wx.BoxSizer(wx.VERTICAL)
            textcontrol1 = wx.TextCtrl(parent, style=wx.TE_PROCESS_ENTER)
            textcontrol1.Bind(wx.EVT_TEXT_ENTER, self.onValueChange)
            inner_sizer.Add(textcontrol1, flag=wx.EXPAND)
            inner_sizer.AddSpacer(5)
            textcontrol2 = wx.TextCtrl(parent, style=wx.TE_PROCESS_ENTER)
            textcontrol2.SetHint('Second value is optional')
            textcontrol2.Bind(wx.EVT_TEXT_ENTER, self.onValueChange)
            inner_sizer.Add(textcontrol2, flag=wx.EXPAND)
            sizer.Add(inner_sizer, flag=wx.EXPAND)
            self.data_widget = [textcontrol1, textcontrol2]
        elif prop.cardinality == Cardinality.ZERO_TO_TWO:  # String or similar, 0-2
            inner_sizer = wx.BoxSizer(wx.VERTICAL)
            textcontrol1 = wx.TextCtrl(parent, style=wx.TE_PROCESS_ENTER)
            textcontrol1.SetHint('Optional')
            textcontrol1.Bind(wx.EVT_TEXT_ENTER, self.onValueChange)
            inner_sizer.Add(textcontrol1, flag=wx.EXPAND)
            inner_sizer.AddSpacer(5)
            textcontrol2 = wx.TextCtrl(parent, style=wx.TE_PROCESS_ENTER)
            textcontrol2.SetHint('Optional')
            textcontrol2.Bind(wx.EVT_TEXT_ENTER, self.onValueChange)
            inner_sizer.Add(textcontrol2, flag=wx.EXPAND)
            sizer.Add(inner_sizer, flag=wx.EXPAND)
            self.data_widget = [textcontrol1, textcontrol2]

    def __setup_string_one_card(self, parent, sizer, prop):
        scroller = wx.lib.scrolledpanel.ScrolledPanel(parent)
        scroll_sizer = wx.BoxSizer(wx.VERTICAL)
        if prop.multiline:
            textcontrol = wx.TextCtrl(scroller, style=wx.TE_MULTILINE)
        else:
            textcontrol = wx.TextCtrl(scroller, style=wx.TE_PROCESS_ENTER)
            textcontrol.Bind(wx.EVT_TEXT_ENTER, self.onValueChange)
        scroll_sizer.Add(textcontrol, flag=wx.EXPAND)
        scroller.Sizer = scroll_sizer
        scroller.SetupScrolling(scroll_x=True, scroll_y=False)
        sizer.Add(scroller, flag=wx.EXPAND)
        # sizer.Add(textcontrol, flag=wx.EXPAND)
        self.data_widget = textcontrol

    def __setup_dropdown(self, parent, sizer, prop):
        if prop.cardinality == Cardinality.ZERO_OR_ONE \
                or prop.cardinality == Cardinality.ONE:
            choice = wx.Choice(parent, size=(400, -1))
            choice.Bind(wx.EVT_CHOICE, lambda e: self.onValueChange(e, False))
            self.data_widget = choice
            self.choice_widget = choice
            sizer.Add(choice, flag=wx.EXPAND)
        if prop.cardinality == Cardinality.ONE_TO_UNBOUND or \
                prop.cardinality == Cardinality.UNBOUND:
            scroller = scrolledPanel.ScrolledPanel(parent)
            inner_sizer = wx.BoxSizer()
            box = wx.ListBox(scroller, size=(400, -1))
            self.data_widget = box
            inner_sizer.Add(box, 1)
            control_sizer = wx.BoxSizer(wx.VERTICAL)
            choice = wx.Choice(scroller, size=(150, -1))
            choice.Bind(wx.EVT_CHOICE,
                        lambda e: parent.add_to_list(e, box, choice,
                                                     choice.GetStringSelection()))
            self.choice_widget = choice
            control_sizer.Add(choice)
            remove_button = wx.Button(scroller, label="Del Selected")
            remove_button.Bind(wx.EVT_BUTTON, lambda event: parent.remove_from_list(event, box))
            control_sizer.Add(remove_button)
            inner_sizer.Add(control_sizer)
            scroller.Sizer = inner_sizer
            scroller.SetupScrolling(scroll_x=True, scroll_y=False)
            sizer.Add(scroller, flag=wx.EXPAND)

    def __setup_address(self, parent, sizer):
        inner_sizer = wx.BoxSizer(wx.VERTICAL)
        text1 = wx.TextCtrl(parent, style=wx.TE_PROCESS_ENTER)
        text1.SetHint('Street')
        text1.Bind(wx.EVT_TEXT_ENTER, self.onValueChange)
        inner_sizer.Add(text1, flag=wx.EXPAND)
        text2 = wx.TextCtrl(parent, style=wx.TE_PROCESS_ENTER)
        text2.SetHint('Postal Code')
        text2.Bind(wx.EVT_TEXT_ENTER, self.onValueChange)
        inner_sizer2 = wx.BoxSizer()
        inner_sizer2.Add(text2, 2)
        inner_sizer2.AddSpacer(5)
        text3 = wx.TextCtrl(parent, style=wx.TE_PROCESS_ENTER)
        text3.SetHint('Locality')
        text3.Bind(wx.EVT_TEXT_ENTER, self.onValueChange)
        inner_sizer2.Add(text3, 10)
        inner_sizer.AddSpacer(5)
        inner_sizer.Add(inner_sizer2, flag=wx.EXPAND)
        sizer.Add(inner_sizer, flag=wx.EXPAND)
        self.data_widget = [text1, text2, text3]

    def __setup_attribution(self, parent, sizer):
        scroller = scrolledPanel.ScrolledPanel(parent)
        inner_sizer = wx.BoxSizer()
        input_sizer = wx.BoxSizer(wx.VERTICAL)
        textcontrol = wx.TextCtrl(scroller, style=wx.TE_PROCESS_ENTER, size=(200, -1))
        textcontrol.SetHint("Role")
        input_sizer.Add(textcontrol)
        choice = wx.Choice(scroller, size=(200, -1))
        choice.SetToolTip("Add a Person or Organization")
        self.choice_widget = choice
        input_sizer.Add(choice)
        inner_sizer.Add(input_sizer)
        inner_sizer.AddSpacer(5)
        button_sizer = wx.BoxSizer(wx.VERTICAL)
        plus_button = wx.Button(scroller, label="+")
        button_sizer.Add(plus_button, flag=wx.EXPAND)
        remove_button = wx.Button(scroller, label="Del Selected")
        button_sizer.Add(remove_button)
        inner_sizer.Add(button_sizer)
        inner_sizer.AddSpacer(5)
        content_list = wx.ListCtrl(scroller, -1, style=wx.LC_REPORT | wx.LC_SINGLE_SEL)
        content_list.InsertColumn(0, 'Role')
        content_list.InsertColumn(1, 'Agent')
        inner_sizer.Add(content_list, 1)
        inner_sizer.Add(wx.Button(scroller))
        scroller.Sizer = inner_sizer
        scroller.SetupScrolling(scroll_x=True, scroll_y=False)
        sizer.Add(scroller, flag=wx.EXPAND)
        self.data_widget = content_list
        remove_button.Bind(wx.EVT_BUTTON,
                           lambda event: parent.remove_from_list(event,
                                                                 content_list))
        plus_button.Bind(wx.EVT_BUTTON,
                         lambda e: parent.add_to_list(e,
                                                      content_list,
                                                      (textcontrol, choice),
                                                      (textcontrol.GetValue(), choice.GetStringSelection())))

    @property
    def data_class(self):
        return self.parent.active_dataset

    @property
    def prop(self):
        return self.data_class.get_prop_by_name(self.prop_name)

    def update_data(self):
        self.prop.value = self.get_value()

    def get_value(self):
        """
        Returns the new property value that has been entered to the UI
        """
        datatype = self.prop.datatype
        cardinality = self.prop.cardinality
        # String or String/URL etc.
        if datatype == Datatype.STRING \
                or datatype == Datatype.STRING_OR_URL \
                or datatype == Datatype.URL \
                or datatype == Datatype.EMAIL \
                or datatype == Datatype.DOWNLOAD \
                or datatype == Datatype.PLACE:
            if cardinality == Cardinality.ONE \
                    or cardinality == Cardinality.ZERO_OR_ONE:
                return self.data_widget.GetValue()
            if cardinality == Cardinality.ONE_TO_TWO \
                    or cardinality == Cardinality.ZERO_TO_TWO:
                return [self.data_widget[0].GetValue(), self.data_widget[1].GetValue()]
            if cardinality == Cardinality.ONE_TO_UNBOUND \
                    or cardinality == Cardinality.ONE_TO_UNBOUND_ORDERED \
                    or cardinality == Cardinality.UNBOUND:
                return self.data_widget.GetStrings()
        elif datatype == Datatype.DATE:
            if cardinality == Cardinality.ONE \
                    or cardinality == Cardinality.ZERO_OR_ONE:
                return self.data_widget.GetLabel()
        elif datatype == Datatype.PERSON_OR_ORGANIZATION or \
                datatype == Datatype.PERSON or \
                datatype == Datatype.ORGANIZATION:
            if cardinality == Cardinality.ZERO_OR_ONE:
                selection = self.data_widget.GetSelection()
                if selection < 0:
                    selection = 0
                string = self.data_widget.GetString(selection)
                return self.metadataset.get_by_string(string)
            if cardinality == Cardinality.ONE_TO_UNBOUND:
                strs = self.data_widget.GetStrings()
                objs = [self.metadataset.get_by_string(s) for s in strs]
                return objs
            if cardinality == Cardinality.UNBOUND:
                strs = self.data_widget.GetStrings()
                objs = [self.metadataset.get_by_string(s) for s in strs]
                return objs
        elif datatype == Datatype.DATA_MANAGEMENT_PLAN:
            return (
                self.data_widget[0].GetValue(),
                self.data_widget[1].GetValue(),
            )
        elif datatype == Datatype.PROJECT:
            return self.metadataset.project
        elif datatype == Datatype.ADDRESS:
            return (
                self.data_widget[0].GetValue(),
                self.data_widget[1].GetValue(),
                self.data_widget[2].GetValue(),
            )
        elif datatype == Datatype.CONTROLLED_VOCABULARY:
            if cardinality == Cardinality.ONE_TO_UNBOUND:
                return self.data_widget.GetStrings()
            elif cardinality == Cardinality.ONE:
                sel = self.data_widget.GetSelection()
                if sel >= 0:
                    s = self.data_widget.GetString(sel)
                    if s in self.prop.value_options:
                        return s
                    else:
                        return ""
                else:
                    return ""
        elif datatype == Datatype.GRANT:
            if cardinality == Cardinality.UNBOUND:
                strs = self.data_widget.GetStrings()
                objs = [self.metadataset.get_by_string(s) for s in strs]
                return objs
        elif datatype == Datatype.ATTRIBUTION:
            if cardinality == Cardinality.ONE_TO_UNBOUND:
                w = self.data_widget
                res = []
                for i in range(w.GetItemCount()):
                    t = w.GetItem(i, 0).GetText()
                    o = self.metadataset.get_by_string(w.GetItem(i, 1).GetText())
                    if t and not t.isspace() and o:
                        res.append((t, o,))
                return res
        elif datatype == Datatype.SHORTCODE:
            return self.data_widget.GetLabel()
        print(f"Couldn't find a type here... weird... {datatype}")
        return "Couldn't find my value... sorry"

    def refresh_ui(self):
        self.set_value(self.prop.value)
        if self.choice_widget:
            self.refresh_choice()
        self.validate()

    def refresh_choice(self):
        options = []
        if self.prop.datatype == Datatype.GRANT:
            options = self.metadataset.grants
        elif self.prop.datatype == Datatype.PERSON:
            options = self.metadataset.persons
        elif self.prop.datatype == Datatype.CONTROLLED_VOCABULARY:
            options = self.prop.value_options
        elif self.prop.datatype == Datatype.ORGANIZATION:
            options = self.metadataset.organizations
        elif self.prop.datatype == Datatype.PERSON_OR_ORGANIZATION or \
                self.prop.datatype == Datatype.ATTRIBUTION:
            options = self.metadataset.persons + self.metadataset.organizations
        options_strs = ["Select to add"] + [str(o) for o in options]
        self.choice_widget.SetItems(options_strs)
        if self.choice_widget == self.data_widget:
            self.set_value(self.prop.value)

    def validate(self):
        widget = self.validity_widget
        res, msg = self.prop.validate()
        if res == Validity.VALID:
            widget.SetForegroundColour(wx.Colour(30, 170, 30))
        elif res == Validity.INVALID_VALUE:
            widget.SetForegroundColour(wx.Colour(170, 30, 30))
        elif res == Validity.REQUIRED_VALUE_MISSING:
            widget.SetForegroundColour(wx.Colour(170, 30, 30))
        elif res == Validity.OPTIONAL_VALUE_MISSING:
            widget.SetForegroundColour(wx.NullColour)
        self.validity_msg = msg

    def set_value(self, val):
        datatype = self.prop.datatype
        cardinality = self.prop.cardinality
        undefined = False
        if not val:
            undefined = True
            if cardinality == Cardinality.ONE \
                    or cardinality == Cardinality.ZERO_OR_ONE:
                val = ""
            elif cardinality == Cardinality.ONE_TO_TWO \
                    or cardinality == Cardinality.ZERO_TO_TWO:
                val = ("", "",)
            else:
                val = []
        # String or String/URL etc.
        if datatype == Datatype.STRING \
                or datatype == Datatype.STRING_OR_URL \
                or datatype == Datatype.URL \
                or datatype == Datatype.EMAIL \
                or datatype == Datatype.DOWNLOAD \
                or datatype == Datatype.PLACE:
            if cardinality == Cardinality.ONE \
                    or cardinality == Cardinality.ZERO_OR_ONE:
                self.data_widget.SetValue(val)
            if cardinality == Cardinality.ONE_TO_TWO \
                    or cardinality == Cardinality.ZERO_TO_TWO:
                self.data_widget[0].SetValue(val[0])
                self.data_widget[1].SetValue(val[1])
            if cardinality == Cardinality.ONE_TO_UNBOUND \
                    or cardinality == Cardinality.ONE_TO_UNBOUND_ORDERED \
                    or cardinality == Cardinality.UNBOUND:
                self.data_widget.SetItems(val)
        elif datatype == Datatype.DATE:
            if cardinality == Cardinality.ONE \
                    or cardinality == Cardinality.ZERO_OR_ONE:
                self.data_widget.SetLabel(val)
        elif datatype == Datatype.PROJECT:
            self.data_widget.SetLabel(str(val))
        elif datatype == Datatype.CONTROLLED_VOCABULARY or \
                datatype == Datatype.GRANT or \
                datatype == Datatype.PERSON_OR_ORGANIZATION or \
                datatype == Datatype.PERSON or \
                datatype == Datatype.ORGANIZATION:
            if cardinality == Cardinality.ZERO_OR_ONE \
                    or cardinality == Cardinality.ONE:
                self.data_widget.SetSelection(
                    self.data_widget.FindString(str(val)))
            if cardinality == Cardinality.ONE_TO_UNBOUND:
                self.data_widget.SetItems([str(v) for v in val])
            if cardinality == Cardinality.UNBOUND:
                self.data_widget.SetItems([str(v) for v in val])
        elif datatype == Datatype.DATA_MANAGEMENT_PLAN:
            if undefined:
                val = (False, "",)
            self.data_widget[0].SetValue(val[0])
            self.data_widget[1].SetValue(val[1])
        elif datatype == Datatype.ADDRESS:
            if undefined:
                val = ("", "", "",)
            self.data_widget[0].SetValue(val[0])
            self.data_widget[1].SetValue(val[1])
            self.data_widget[2].SetValue(val[2])
        elif datatype == Datatype.ATTRIBUTION:
            if undefined:
                val = [("", "",)]
            self.data_widget.DeleteAllItems()
            for v in val:
                self.data_widget.Append((v[0], str(v[1])))
                self.data_widget.SetColumnWidth(0, -1)
                self.data_widget.SetColumnWidth(1, -1)
        elif datatype == Datatype.SHORTCODE:
            return self.data_widget.SetLabel(val)

    def onValueChange(self, event, navigate: bool = True):
        data_handler.update_all()
        if navigate:
            event.GetEventObject().Navigate()


class DataTab(scrolledPanel.ScrolledPanel):

    def __init__(self, parent, metadataset, dataset, title, multiple=False):
        """
        The class does the following: ToDo: Check the params, dataclass?

        Args:
            parent (object): to be checked
            metadataset (object): to be checked
            dataset (object): to be checked
            title (str): title of the project
            multiple (bool): false
        """
        # wx.Panel.__init__(self, parent, style=wx.EXPAND)
        scrolledPanel.ScrolledPanel.__init__(self, parent, -1)

        self.parent = parent
        self.dataset = dataset
        self.multiple = multiple
        self.metadataset = metadataset
        self.rows = []
        self.multiple_selection = 0
        outer_sizer = wx.BoxSizer(wx.VERTICAL)
        sizer = wx.FlexGridSizer(4, 10, 10)

        if dataset:
            ds = dataset
            if multiple:
                ds = self.active_dataset
            for i, prop in enumerate(ds.get_properties()):
                row = PropertyRow(self, prop, sizer, i, metadataset)
                self.rows.append(row)

        sizer.AddGrowableCol(1)

        if multiple:
            dataset_sizer = wx.BoxSizer()
            dataset_listbox = wx.ListBox(self, size=(200, -1))
            for ds in dataset:
                dataset_listbox.Append(str(ds))
            dataset_listbox.Bind(
                wx.EVT_LISTBOX, lambda e: self.change_selection(e))
            dataset_listbox.Select(0)
            self.multiple_listbox = dataset_listbox
            dataset_sizer.Add(dataset_listbox, wx.EXPAND)
            dataset_sizer.AddSpacer(5)
            button_sizer = wx.BoxSizer(wx.VERTICAL)
            button_add = wx.Button(self, label="Add New")
            button_add.Bind(wx.EVT_BUTTON, lambda e: self.add_object(
                e, dataset_listbox, title))
            button_sizer.Add(button_add, flag=wx.EXPAND)
            button_remove = wx.Button(self, label="Remove Selected")
            button_remove.Bind(
                wx.EVT_BUTTON, lambda e: self.remove_object(e, dataset_listbox))
            button_sizer.Add(button_remove)
            dataset_sizer.Add(button_sizer)
            outer_sizer.Add(dataset_sizer, flag=wx.EXPAND)
            outer_sizer.AddSpacer(20)
        outer_sizer.Add(sizer, flag=wx.EXPAND)
        self.SetSizer(outer_sizer)
        self.SetupScrolling(scroll_x=False, scroll_y=True)
        self.Fit()
        self.Layout()
        # self.SetScrollbars(0, 16, 60, 15)

    @property
    def active_dataset(self):
        if self.multiple:
            return self.dataset[self.multiple_selection]
        else:
            return self.dataset

    def update_data(self):
        for row in self.rows:
            row.update_data()

    def refresh_ui(self):
        if self.multiple:
            self.multiple_listbox.SetItems([str(ds) for ds in self.dataset])
            if self.multiple_selection < 0:
                self.multiple_selection = len(self.multiple_listbox.GetItems()) - 1
            self.multiple_listbox.SetSelection(self.multiple_selection)
        for row in self.rows:
            row.refresh_ui()
        self.Layout()

    def add_object(self, event, listbox, title):
        if title == "Person":
            self.metadataset.add_person()
        if title == "Dataset":
            self.metadataset.add_dataset()
        elif title == "Organization":
            self.metadataset.add_organization()
        elif title == "Grant":
            self.metadataset.add_grant()
        self.multiple_selection = -1
        data_handler.refresh_ui()

    def remove_object(self, event, listbox):
        selection = listbox.GetSelection()
        if selection >= 0 and listbox.GetCount() > 1:
            s = listbox.GetStringSelection()
            self.metadataset.remove(self.metadataset.get_by_string(s))
            self.multiple_selection = selection - 1
        data_handler.update_all()
        data_handler.refresh_ui()

    def change_selection(self, event):
        sel = event.GetEventObject().GetSelection()
        data_handler.update_all()
        self.multiple_selection = sel
        data_handler.refresh_ui()

    def add_to_list(self, event, content_list, widget, addable):
        """
        add an object to a listbox.
        """
        if not addable:  # is None
            return
        if isinstance(widget, tuple):  # Attribution, i.e. two inputs
            role = addable[0]
            agent = addable[1]
            if not role or not agent or \
                    role.isspace() or agent.isspace() or \
                    agent == "Select to add":
                print('invalid input')
                return
            for i in range(content_list.GetItemCount()):
                r = content_list.GetItem(i, 0).GetText()
                a = content_list.GetItem(i, 1).GetText()
                if r == role and a == agent:
                    print('Item already exists')
                    return
            content_list.Append((role, agent))
        else:
            if str(addable).isspace() or \
                    addable == "Select to add" or \
                    str(addable) in content_list.GetStrings():
                self.reset_widget(widget)
                return
            content_list.Append(str(addable))
            self.reset_widget(widget)
        data_handler.update_all()

    def reset_widget(self, widget):
        if isinstance(widget, wx.StaticText) or \
                isinstance(widget, wx.TextCtrl):
            widget.SetValue('')
        elif isinstance(widget, wx.Choice):
            widget.SetSelection(0)

    def remove_from_list(self, event, content_list):
        """
        remove an object from a listbox.
        """
        if isinstance(content_list, wx.ListCtrl):
            selection = content_list.GetFirstSelected()
            if selection >= 0:
                content_list.DeleteItem(selection)
        else:
            selection = content_list.GetSelection()
            if selection >= 0:
                content_list.Delete(selection)
        data_handler.update_all()

    def show_help(self, evt, message, sample):
        """
        Show a help dialog
        """
        msg = f"Description:\n{message}\n\nExample:\n{sample}"
        win = HelpPopup(self, msg)
        btn = evt.GetEventObject()
        pos = btn.ClientToScreen((0, 0))
        sz = btn.GetSize()
        win.Position(pos, (0, sz[1]))
        win.Popup()

    def show_validity(self, evt, val, card):
        """
        Show a help dialog
        """
        msg = f"Validation Result:\n{val}\n\nExpected Cardinality:\n{card}"
        win = HelpPopup(self, msg)
        btn = evt.GetEventObject()
        pos = btn.ClientToScreen((0, 0))
        sz = btn.GetSize()
        win.Position(pos, (0, sz[1]))
        win.Popup()

    def pick_date(self, evt, label: wx.StaticText, prop):
        with CalendarDlg(self, prop.name, label.GetLabel()) as dlg:
            if dlg.ShowModal() == wx.ID_OK:
                label.SetLabel(dlg.cal.Date.FormatISODate())
                data_handler.update_all()


class HelpPopup(wx.PopupTransientWindow):
    def __init__(self, parent, msg):
        """
        This class provides a help message
        Args:
            parent (object): the parent class
            msg (str): the help text with examples
        """
        wx.PopupTransientWindow.__init__(self, parent)
        panel = wx.Panel(self)
        st = wx.StaticText(panel, -1, msg)
        sizer = wx.BoxSizer(wx.VERTICAL)
        sizer.Add(st, 0, wx.ALL, 5)
        panel.SetSizer(sizer)
        sizer.Fit(panel)
        sizer.Fit(self)
        self.Layout()


class TabbedWindow(wx.Frame):
    def __init__(self, parent, dataset: MetaDataSet):
        """
        This class organizes the different tabs
        ToDo: What exactly is the metadataset?

        Args:
            parent (object): the parent object
            dataset (object): the dataset
            MetaDataSet (object): the metadataset
        """
        wx.Frame.__init__(self, parent, id=-1, title="", pos=wx.DefaultPosition,
                          size=(1200, 600), style=wx.DEFAULT_FRAME_STYLE,
                          name="Metadata tabs")
        self.Bind(wx.EVT_CLOSE, self.on_close)
        self.panel = wx.Panel(self)
        self.parent = parent
        self.dataset = dataset

        # Create a panel and notebook (tabs holder)
        panel = self.panel
        nb = wx.Notebook(panel)
        nb.SetMinSize((-1, 200))
        nb.Bind(wx.EVT_NOTEBOOK_PAGE_CHANGED, self.on_tab_change)

        # Create the tab windows
        tab1 = TabOne(nb, self.dataset)
        tab2 = DataTab(nb, self.dataset, self.dataset.project, "Project")
        tab3 = DataTab(nb, self.dataset, self.dataset.dataset, "Dataset", multiple=True)
        tab4 = DataTab(nb, self.dataset, self.dataset.persons, "Person", multiple=True)
        tab5 = DataTab(nb, self.dataset, self.dataset.organizations, "Organization", multiple=True)
        tab6 = DataTab(nb, self.dataset, self.dataset.grants, "Grant", multiple=True)

        # Add the windows to tabs and name them.
        nb.AddPage(tab1, "Base Data")
        nb.AddPage(tab5, "Organization")
        nb.AddPage(tab4, "Person")
        nb.AddPage(tab6, "Grant")
        nb.AddPage(tab3, "Dataset")
        nb.AddPage(tab2, "Project")

        data_handler.tabs = [tab2, tab3, tab4, tab5, tab6]

        nb_sizer = wx.BoxSizer()
        nb_sizer.Add(nb, 1, wx.ALL | wx.EXPAND)

        # Buttons
        save_button = wx.Button(panel, label='Save')
        save_button.Bind(wx.EVT_BUTTON, self.on_save)
        saveclose_button = wx.Button(panel, label='Save and Close')
        saveclose_button.Bind(wx.EVT_BUTTON, self.on_saveclose)
        cancel_button = wx.Button(panel, label='Cancel')
        cancel_button.Bind(wx.EVT_BUTTON, self.on_close)
        button_sizer = wx.BoxSizer(wx.HORIZONTAL)
        button_sizer.Add(save_button, 0, wx.ALL, 5)
        button_sizer.Add(saveclose_button, 0, wx.ALL, 5)
        button_sizer.Add(cancel_button, 0, wx.ALL, 5)
        feedback_text = wx.StaticText(panel)
        self.feedback_text = feedback_text
        button_sizer.Add(feedback_text, 0, wx.ALL, 5)

        # Set notebook in a sizer to create the layout
        sizer = wx.FlexGridSizer(1, 2, 10)
        sizer.AddGrowableCol(0)
        sizer.AddGrowableRow(0)
        sizer.Add(nb_sizer, 0, wx.ALL | wx.EXPAND, 5)
        sizer.Add(button_sizer, 0, wx.ALL | wx.ALIGN_BOTTOM, 5)
        panel.SetSizer(sizer)
        # sizer.Fit(self)

    def on_tab_change(self, event):
        data_handler.update_all()

    def on_save(self, event):
        self.save()
        self.feedback("Saved Successfully.")

    def on_saveclose(self, event):
        self.save()
        self.close()

    def on_close(self, event):
        self.close()

    def save(self):
        data_handler.update_all()
        # data_handler.validate_graph(self.dataset)  # TODO: why was that here again?
        data_handler.save_data(dataset=self.dataset)
        data_handler.refresh_ui()

    def close(self):
        self.parent.Enable()
        self.parent.refresh_view()
        data_handler.current_window = None
        self.Destroy()

    def get_persons(self):
        return self.dataset.persons

    def get_organizations(self):
        return self.dataset.organizations

    def feedback(self, msg, success=True):
        if success:
            self.feedback_text.SetForegroundColour(wx.Colour(50, 200, 50))
        else:
            self.feedback_text.SetForegroundColour(wx.Colour(200, 50, 50))
        self.feedback_text.SetLabel(msg)
        wx.CallLater(2500, lambda: self.feedback_text.SetLabel(''))


if __name__ == '__main__':
    collectMetadata()



import sys
import wx
import wx.adv


class CalendarDlg(wx.Dialog):
    # def __init__(self, parent):
    def __init__(self, parent, title, date_str):

        # wx.Dialog.__init__(self, parent, title=parent.title)
        wx.Dialog.__init__(self, parent, title=title)
        panel = wx.Panel(self, -1)

        sizer = wx.BoxSizer(wx.VERTICAL)
        panel.SetSizer(sizer)

        date = wx.DateTime()
        date.ParseDate(date_str)
        cal = wx.adv.GenericCalendarCtrl(panel, date=date)
        # cal = wx.adv.GenericCalendarCtrl(panel, date=parent.date)

        if sys.platform != 'win32':
            # gtk truncates the year - this fixes it
            w, h = cal.Size
            cal.Size = (w + 15, h + 85)
            cal.MinSize = cal.Size

        sizer.Add(cal, 0)

        button_sizer = wx.BoxSizer(wx.HORIZONTAL)
        button_sizer.Add((0, 0), 1)
        btn_ok = wx.Button(panel, wx.ID_OK)
        btn_ok.SetDefault()
        button_sizer.Add(btn_ok, 0, wx.ALL, 2)
        button_sizer.Add((0, 0), 1)
        btn_can = wx.Button(panel, wx.ID_CANCEL)
        button_sizer.Add(btn_can, 0, wx.ALL, 2)
        button_sizer.Add((0, 0), 1)
        sizer.Add(button_sizer, 1, wx.EXPAND | wx.ALL, 10)
        sizer.Fit(panel)
        self.ClientSize = panel.Size

        cal.Bind(wx.EVT_KEY_DOWN, self.on_key_down)
        cal.SetFocus()
        self.cal = cal

    def on_key_down(self, evt):
        code = evt.KeyCode
        if code == wx.WXK_TAB:
            self.cal.Navigate()
        elif code in (wx.WXK_RETURN, wx.WXK_NUMPAD_ENTER):
            self.EndModal(wx.ID_OK)
        elif code == wx.WXK_ESCAPE:
            self.EndModal(wx.ID_CANCEL)
        else:
            evt.Skip()



import os
import pickle
import shutil

from .metaDataSet import MetaDataSet
from .utils import open_file


class DataHandling:
    """ This class handles data.

    It checks for availability in the filesystem, and
    if not, creates the data structure. It also takes care for the storage on disk.
    The data are stored as a pickle file.

    The class should not be called as a static.
    Rather, there should at any given time be an instance of this class (`data_handler`) be available.
    All calls should be done on this instance, as it holds the actual data representation.
    """

    def __init__(self):
        self.current_window = None
        self.projects = []
        self.tabs = []
        self.data_storage = os.path.expanduser("~") + "/DaSCH/config/repos.data"
        # LATER: path could be made customizable
        self.load_data()

    def add_project(self, folder_path: str, shortcode: str, files: list):
        """
        Add a new project.

        This project adds a new project folder to the collection after the user specified the folder.

        The Project is appended at the end of the list.

        Args:
            folder_path (str): path to the project folder
            shortcode (str): the project shortcode
            files (list): the files in the project folder
        """
        folder_name = os.path.basename(folder_path)
        dataset = MetaDataSet(folder_name, folder_path, shortcode)
        self.projects.append(dataset)
        dataset.files += files
        self.save_data()

    def remove_project(self, project):
        if project and project in self.projects:
            self.projects.remove(project)
            self.save_data()

    def load_data(self):
        """
        Load data from previous runtimes (if any).

        Currently, this checks `~/DaSCH/config/repos.data`.
        """
        if not os.path.exists(self.data_storage):
            os.makedirs(os.path.dirname(self.data_storage), exist_ok=True)
            return
        with open(self.data_storage, 'rb') as file:
            self.projects = pickle.load(file)

    def save_data(self, dataset: MetaDataSet = None):
        """
        Save data to disc.

        Currently, the data are stored under `~/DaSCH/config/repos.data`.
        """
        # LATER: could let the user decide where to store the data.
        # TODO: export metadata here
        if dataset:
            dataset.generate_rdf_graph()
        with open(self.data_storage, 'wb') as file:
            pickle.dump(self.projects, file)

    def validate_and_export_data(self, index: int) -> tuple:
        project = self.projects[index]
        validation_result = self.validate_graph(project)
        try:
            graph = project.generate_rdf_graph()
            if not graph:
                raise Exception('No Graph')
        except Exception:
            print('Warning: could not load graph from cache. Performance may be decreased.')
            # LATER: remove with next breaking change
            graph = project.graph
        self.export_rdf(project.path, graph)
        return validation_result

    def import_project(self, path):
        try:
            with open(path, 'rb') as f:
                dataset = pickle.load(f)
                self.projects.append(dataset)
        except Exception:
            import traceback
            traceback.print_exc()
            print(f'\n\n--------\n\nCould not import file: {path}')

    def export_rdf(self, path, graph, show=True):
        path += '/metadata'
        if not os.path.exists(path):
            os.makedirs(path)
        p = path + '/metadata.ttl'
        with open(p, 'w') as f:
            s = graph.serialize(format='turtle').decode("utf-8")
            f.write(s)
        p = path + '/metadata.json'
        with open(p, 'w') as f:
            s = graph.serialize(format='json-ld').decode("utf-8")
            f.write(s)
        p = path + '/metadata.xml'
        with open(p, 'w') as f:
            s = graph.serialize(format='xml').decode("utf-8")
            f.write(s)
        if show:
            open_file(path)

    def zip_and_export(self, dataset: MetaDataSet, target: str):
        if not dataset:
            return
        if not target:
            target = dataset.path
        target_file = os.path.join(target, dataset.name)
        try:
            graph = dataset.generate_rdf_graph()
            if not graph:
                raise Exception('No Graph')
        except Exception:
            print('Warning: could not load graph from cache. Performance may be decreased.')
            # LATER: remove with next breaking change
            graph = dataset.graph
        self.export_rdf(dataset.path, graph)
        p = dataset.path
        tmp = os.path.join(p, '.tmp')
        meta = os.path.join(p, 'metadata')
        os.makedirs(tmp, exist_ok=True)
        tmp_m = os.path.join(tmp, 'metadata')
        os.makedirs(tmp_m, exist_ok=True)
        pickle_path = os.path.join(tmp, 'binary')
        os.makedirs(pickle_path, exist_ok=True)
        for f in dataset.files:
            shutil.copy(os.path.join(p, f), tmp)
        shutil.copytree(meta, tmp_m, dirs_exist_ok=True)
        shortcode = dataset.shortcode
        with open(os.path.join(pickle_path, f'project_{shortcode}.data'), mode='wb') as pick:
            pickle.dump(dataset, pick)
        shutil.make_archive(target_file, 'zip', tmp)
        shutil.rmtree(tmp, ignore_errors=True)
        open_file(target)

    def validate_graph(self, dataset: MetaDataSet) -> tuple:
        """
        Validates all properties in a specific `MetaDataSet.`

        Does not validate each of the properties separately,
        but rather generates the RDF graph, which then gets validated.
        """
        try:
            graph = dataset.generate_rdf_graph()
            if not graph:
                raise Exception('No Graph')
        except Exception:
            print('Warning: could not load graph from cache. Performance may be decreased.')
            # LATER: remove with next breaking change
            graph = dataset.graph
        return dataset.validate_graph(graph)

    def update_all(self):
        """
        Update date from GUI.

        Calling this function iterates over each Property in the dataset
        and updates it with the value found in its corresponding GUI component.
        """
        for tab in self.tabs:
            tab.update_data()
        self.refresh_ui()

    def refresh_ui(self):
        """
        Refresh all values in the UI according to the saved values.

        This method also invokes on-the-fly validation.
        Note: Calling this method discards all changes that have not been updated in the metaDataSet.
        """
        for tab in self.tabs:
            tab.refresh_ui()

    def get_project_by_shortcode(self, shortcode):
        for p in self.projects:
            if p.shortcode == shortcode:
                return p



import os
import re
import platform
import subprocess
import validators
from enum import Enum
from urllib.parse import urlparse


def open_file(path):
    if platform.system() == "Windows":
        os.startfile(path)
    elif platform.system() == "Darwin":
        subprocess.Popen(["open", path])
    else:
        subprocess.Popen(["xdg-open", path])


def areURLs(urls: list):
    for url in urls:
        if not url:
            continue
        if not isURL(url):
            return False
    return True


def isURL(url: str):
    if url and not url.isspace():
        if validators.url(url):
            return True
        if validators.url('http://' + url):
            return True
        # LATER: good enough?
        # if validators.url('http://www.' + url):
        #     return True
    return False


def get_url_property_id(url: str) -> str:
    """
    This method tries to guess the propetyID for a URL.

    For certain pre-defined cases, a reasonable propertyID is chosen;
    otherwise, the net location is being extracted, if possible.

    Args:
        url (str): a URL

    Returns:
        str: a propertyID
    """
    if re.search(r'skos\.um\.es', url):
        return "SKOS UNESCO Nomenclature"
    if re.search(r'geonames\.org', url):
        return "Geonames"
    if re.search(r'pleiades\.stoa\.org', url):
        return "Pleiades"
    if re.search(r'orcid\.org', url):
        return "ORCID"
    if re.search(r'viaf\.org', url):
        return "VIAF"
    if re.search(r'\/gnd\/', url) or re.search(r'portal\.dnb\.de', url):
        return "GND"
    if re.search(r'n2t\.net\/ark\:\/99152', url):
        return "Periodo"
    if re.search(r'chronontology\.dainst\.org', url):
        return "ChronOntology"
    if re.search(r'creativecommons\.', url):
        return "Creative Commons"
    # LATER: propertyID's for common institutions
    loc = urlparse(url).netloc
    if len(loc.split('.')) > 2:
        return '.'.join(loc.split('.')[1:])
    if loc:
        return loc
    return url[:12]


def are_emails(mails: list):
    for mail in mails:
        if not mail:
            continue
        if not is_email(mail):
            return False
    return True


def is_email(mail: str):
    if mail and not mail.isspace():
        if validators.email(mail):
            return True
    return False


class Validity(Enum):
    VALID = 0
    INVALID_VALUE = 1
    REQUIRED_VALUE_MISSING = 2
    OPTIONAL_VALUE_MISSING = 3


class Cardinality(Enum):
    """
    A set of cardinalities that may be used for properties.
    """
    UNBOUND = 0
    ONE = 1
    ZERO_OR_ONE = 2
    ONE_TO_UNBOUND = 3
    ONE_TO_TWO = 4
    ZERO_TO_TWO = 5
    ONE_TO_UNBOUND_ORDERED = 6

    def get_optionality_string(card) -> str:
        """
        Returns wether or not a cardinality is optional.

        Args:
            card (Cardinality): the cardinality in question

        Returns:
            str: "Mandatory" or "Optional", depending on the cardinality
        """
        if Cardinality.isMandatory(card):
            return "Mandatory"
        else:
            return "Optional"

    def isMandatory(card) -> bool:
        if card == Cardinality.ONE \
                or card == Cardinality.ONE_TO_TWO \
                or card == Cardinality.ONE_TO_UNBOUND \
                or card == Cardinality.ONE_TO_UNBOUND_ORDERED:
            return True
        if card == Cardinality.UNBOUND \
                or card == Cardinality.ZERO_OR_ONE \
                or card == Cardinality.ZERO_TO_TWO:
            return False

    def as_sting(card) -> str:
        if card == Cardinality.UNBOUND:
            return "Unbound: 0-n values"
        elif card == Cardinality.ONE:
            return "Exactly one value"
        elif card == Cardinality.ZERO_OR_ONE:
            return "Optional: Zero or one value"
        elif card == Cardinality.ONE_TO_UNBOUND:
            return "Mandatory unbound: 1-n values"
        elif card == Cardinality.ONE_TO_TWO:
            return "One or two values"
        elif card == Cardinality.ZERO_TO_TWO:
            return "Optional: Zero, one or two values"
        elif card == Cardinality.ONE_TO_UNBOUND_ORDERED:
            return "Mandatory unbound: 1-n values (ordered)"


class Datatype(Enum):
    """
    A set of cardinalities that may be used for properties.
    """
    STRING = 0
    DATE = 1
    STRING_OR_URL = 2
    PLACE = 3
    PERSON_OR_ORGANIZATION = 4
    GRANT = 5
    DATA_MANAGEMENT_PLAN = 6
    URL = 7
    CONTROLLED_VOCABULARY = 8
    PROJECT = 9
    ATTRIBUTION = 10
    EMAIL = 11
    ADDRESS = 12
    PERSON = 13
    ORGANIZATION = 14
    DOWNLOAD = 15
    SHORTCODE = 16



"""
This module holds the data structure that is used for modelling metadata.

The classes defined here aim to represent a metadata-set, closely following the metadata ontology.
"""

from abc import ABC, abstractmethod
import re
import pyshacl
import validators
from rdflib import Graph, URIRef, RDF, Literal, Namespace, BNode
from rdflib.namespace import SDO, XSD

from . import utils
from .utils import Cardinality, Datatype, Validity


ontology_url = "https://raw.githubusercontent.com/dasch-swiss/dsp-ontologies/main/dsp-repository/v1/dsp-repository.shacl.ttl"
dsp_repo = Namespace("http://ns.dasch.swiss/repository#")
prov = Namespace("http://www.w3.org/ns/prov#")


class MetaDataSet:
    """ Representation of a data set.

    This class represents a data set of project metadata.
    It holds the following properties:
    - name: the repo/project name.
      Typically the name of the folder that was selected.
    - path: the full path of the folder that was selected.
    - files: a list of relevant files in the folder
    - project: a `metaDataSet.Project` representation of the actual metadata (as specified by the ontology).
    - dataset: a list of `metaDataSet.Dataset`
    - persons: a list of `metaDataSet.Person`
    - organizations: a list of `metaDataSet.Organization`

    At a later stage, this class should be able to return a representation of its data in form of an RDF graph.
    """

    @property
    def name(self):
        return self.__name

    @name.setter
    def name(self, name: str):
        self.__name = name

    @property
    def path(self):
        return self.__path

    @path.setter
    def path(self, path: str):
        self.__path = path

    @property
    def files(self):
        return self.__files

    @files.setter
    def files(self, files: list):
        self.__files = files

    @property
    def graph(self):
        return self.__graph

    @graph.setter
    def graph(self, graph: Graph):
        self.__graph = graph
        ttl = graph.serialize(format='turtle').decode("utf-8")
        self.turtle = ttl
        val = self.validate_graph(graph)[0]
        self.validation_result = val

    @property
    def turtle(self):
        return self.__turtle

    @turtle.setter
    def turtle(self, turtle: str):
        self.__turtle = turtle

    @property
    def validation_result(self):
        return self.__validity

    @validation_result.setter
    def validation_result(self, validity: bool):
        self.__validity = validity

    def __init__(self, name: str, path: str, shortcode: str):
        """
        Initiates the object.

        Args:
            name (str): [description]
            path (str): [description]
            shortcode (str): [description]
        """
        self.shortcode = shortcode
        self.name = name
        self.path = path
        self.files = []
        self.project = Project(name, shortcode, self)
        self.dataset = [Dataset(name, self.project, self)]
        self.persons = [Person(self)]
        self.organizations = [Organization(self)]
        self.grants = [Grant(self)]
        self.update_iris()

    def update_iris(self):
        """
        Updates the IRIs of all DataClass fields.

        This method should be called whenever something is added/removed
        to the lists holding DataClass instances (persons, etc.).
        """
        # TODO: this method needs to be called whenever a person/org/dataset is added/removed
        # TODO: ensure updating the IRIs doesn't mess up identifying objects by string
        self.project.iri_suffix = "-project"
        for i, ds in enumerate(self.dataset):
            ds.iri_suffix = f"-dataset-{str(i + 1).zfill(3)}"
        for i, person in enumerate(self.persons):
            person.iri_suffix = f"-person-{str(i + 1).zfill(3)}"
        for i, org in enumerate(self.organizations):
            org.iri_suffix = f"-organization-{str(i + 1).zfill(3)}"
        for i, g in enumerate(self.grants):
            g.iri_suffix = f"-grant-{str(i + 1).zfill(3)}"

    def __str__(self):
        return str({
            "name": self.name,
            "path": self.path,
            "files": self.files,
            "metadata": [
                self.project,
                self.dataset,
                self.persons,
                self.organizations,
                self.grants,
            ]
        })

    def get_all_properties(self) -> list:
        """
        Returns a list of all properties held by fields of this class. (person, dataset, etc.)
        """
        res = self.project.get_properties()
        for p in self.dataset:
            res.extend(p.get_properties())
        for p in self.persons:
            res.extend(p.get_properties())
        for o in self.organizations:
            res.extend(o.get_properties())
        for o in self.grants:
            res.extend(o.get_properties())
        return res

    def validate_graph(self, graph) -> tuple:
        """
        Validates the graph of the entire data against the SHACL ontology.
        """
        conforms, results_graph, results_text = pyshacl.validate(graph, shacl_graph=ontology_url)
        return conforms, results_graph, results_text

    def generate_rdf_graph(self) -> Graph:
        """
        Generates the RFD graph of the entire dataset.

        Returns:
            Graph: The RDF graph
        """
        graph = Graph(base=dsp_repo)
        graph.bind("dsp-repo", dsp_repo)
        graph.bind("schema", SDO)
        graph.bind("xsd", XSD)
        graph.bind("prov", prov)
        self.project.add_rdf_to_graph(graph, "Project")
        for i, person in enumerate(self.dataset):
            person.add_rdf_to_graph(graph, "Dataset")
        for i, person in enumerate(self.persons):
            person.add_rdf_to_graph(graph, "Person")
        for i, org in enumerate(self.organizations):
            org.add_rdf_to_graph(graph, "Organization")
        for i, org in enumerate(self.grants):
            org.add_rdf_to_graph(graph, "Grant")
        try:
            self.graph = graph
        except Exception:
            print('Warning: Graph could not be cached.')
            # LATER: remove with next breaking change
        return graph

    def get_by_string(self, s: str):
        if not s or s.isspace():
            return
        if str(self.project) == s:
            return self.project
        id_str = s.split(':')[0]
        for o in self.dataset:
            if str(o).startswith(id_str):
                return o
        for o in self.persons:
            if str(o).startswith(id_str):
                return o
        for o in self.organizations:
            if str(o).startswith(id_str):
                return o
        for o in self.grants:
            if str(o).startswith(id_str):
                return o

    def add_dataset(self):
        new = Dataset(self.name, self.project, self)
        self.dataset.append(new)
        self.update_iris()

    def add_person(self):
        new = Person(self)
        self.persons.append(new)
        self.update_iris()

    def add_organization(self):
        new = Organization(self)
        self.organizations.append(new)
        self.update_iris()

    def add_grant(self):
        new = Grant(self)
        self.grants.append(new)
        self.update_iris()

    def remove(self, obj):
        if obj in self.dataset:
            self.dataset.remove(obj)
        if obj in self.persons:
            self.persons.remove(obj)
        if obj in self.organizations:
            self.organizations.remove(obj)
        if obj in self.grants:
            self.grants.remove(obj)

    def get_status(self):
        try:
            g = self.graph
            if not g:
                raise Exception('no graph')
        except Exception:
            print('Warning: Could not load cached graph. Performance may be decreased')
            # LATER: remove with next breaking change
            g = self.generate_rdf_graph()
        try:
            val = self.validation_result
        except Exception:
            # LATER: remove with next breaking change
            print('Warning: no cached validation result available. Performance may be decreased.')
            val = self.validate_graph(g)[0]
        if val:
            overall = 'Valid'
        else:
            overall = 'Invalid'
        invalid = 0
        missing = 0
        optional = 0
        valid = 0
        for p in self.get_all_properties():
            v, _ = p.validate()
            if v == Validity.INVALID_VALUE:
                invalid += 1
            elif v == Validity.REQUIRED_VALUE_MISSING:
                missing += 1
            elif v == Validity.OPTIONAL_VALUE_MISSING:
                optional += 1
            elif v == Validity.VALID:
                valid += 1
        return f"{overall}  --  {invalid + missing} Problems; {valid} Values"

    def get_turtle(self) -> str:
        try:
            ttl = self.turtle
            if not isinstance(ttl, str):
                raise Exception('turtle was not a string')
            if ttl:
                return ttl
            else:
                raise Exception('no turtle found')
        except Exception as e:
            print('Warning: No turtle cached. Performance may be decreased.')
            print(f'    (Exception: {e})')
            # LATER: remove with next breaking change
            g = self.generate_rdf_graph()
            return g.serialize(format='turtle').decode("utf-8")


class DataClass(ABC):
    """
    Abstract parent class of all classes holding data.
    """

    def get_metadataset(self) -> MetaDataSet:
        """
        Returns the `MetaDataSet` to which this class belongs

        Returns:
            MetaDataSet: The owner MetaDataSet
        """
        return self.meta

    def add_rdf_to_graph(self, graph: Graph, typename: str):
        """
        Adds the RDF representation of this class to a graph

        Args:
            graph (Graph): the `rdflib.Graph` to which the triples should be added
            typename (str): the type name of this class
        """
        iri = self.get_rdf_iri()
        type = dsp_repo[typename]
        graph.add((iri, RDF.type, type))
        for prop in self.get_properties():
            graph += prop.get_triples(iri)

    # TODO: ensure that this gets updated whenever the shortcode changes
    def get_rdf_iri(self) -> URIRef:
        """
        Return the iri of the object.

        This can be called to use the object as subject of an RDF triple.

        Returns:
            URIRef: the IRI
        """
        shortcode = self.get_metadataset().project.shortcode.value
        if not shortcode:
            shortcode = "xxxx"
        classname = 'dsp-' + shortcode + self.iri_suffix
        return URIRef(dsp_repo[classname])

    @abstractmethod
    def get_properties(self) -> list:
        """
        Return a list of all `Property` fields of this class
        """
        raise NotImplementedError

    def get_prop_by_name(self, name):
        for p in self.get_properties():
            if p.name == name:
                return p


class Project(DataClass):
    """
    Project shape.

    Corresponds to `dsp-repo:Project` in our ontology.

    Args:
        name (str): The name of the project
        meta (MetaDataSet): the owning `MetaDataSet`
    """

    def __init__(self, name: str, shortcode: str, meta: MetaDataSet):
        self.meta = meta
        self.name = Property(meta, "Name",
                             "The name of the Project",
                             "Test Project",
                             Datatype.STRING,
                             Cardinality.ONE,
                             name,
                             predicate=dsp_repo.hasName)

        msg = "This is a test project. All properties have been used to test these."
        msg += "\nYou will just describe your project briefly."
        self.description = Property(meta, "Description",
                                    "Description of the Project",
                                    msg,
                                    Datatype.STRING,
                                    Cardinality.ONE,
                                    predicate=dsp_repo.hasDescription,
                                    multiline=True)

        msg = "mathematics, science, history of science, history of mathematics."
        msg += "\nUse the plus sign to have a new field for each key word."
        self.keywords = Property(meta, "Keywords",
                                 "Keywords and tags",
                                 msg,
                                 Datatype.STRING,
                                 Cardinality.ONE_TO_UNBOUND,
                                 predicate=dsp_repo.hasKeywords)

        msg = "Discipline and research fields from UNESCO nomenclature: https://skos.um.es/unesco6/?l=en"
        msg += "\nor from http://www.snf.ch/SiteCollectionDocuments/allg_disziplinenliste.pdf"
        self.discipline = Property(meta, "Discipline",
                                   msg,
                                   "http://skos.um.es/unesco6/11",
                                   Datatype.STRING_OR_URL,
                                   Cardinality.ONE_TO_UNBOUND,
                                   predicate=dsp_repo.hasDiscipline)

        self.startDate = Property(meta, "Start Date",
                                  "The date when the project started, e. g. when funding was granted.",
                                  "2000-07-26T21:32:52",
                                  Datatype.DATE,
                                  Cardinality.ONE,
                                  predicate=dsp_repo.hasStartDate)

        msg = "The date when the project was finished, e. g. when the last changes to the project data where completed."
        self.endDate = Property(meta, "End Date",
                                msg,
                                "2000-07-26T21:32:52",
                                Datatype.DATE,
                                Cardinality.ZERO_OR_ONE,
                                predicate=dsp_repo.hasEndDate)

        self.temporalCoverage = Property(meta, "Temporal Coverage",
                                         "Temporal coverage of the project from http://perio.do/en/\nor https://chronontology.dainst.org/",
                                         "http://chronontology.dainst.org/period/Ef9SyESSafJ1",
                                         Datatype.STRING_OR_URL,
                                         Cardinality.ONE_TO_UNBOUND,
                                         predicate=dsp_repo.hasTemporalCoverage)

        self.spatialCoverage = Property(meta, "Spatial Coverage",
                                        "Spatial coverage of the project from Geonames URL: https://www.geonames.org/\nor from Pleiades URL: https://pleiades.stoa.org/places",
                                        "https://www.geonames.org/6255148/europe.html",
                                        Datatype.PLACE,
                                        Cardinality.ONE_TO_UNBOUND,
                                        predicate=dsp_repo.hasSpatialCoverage)

        self.funder = Property(meta, "Funder",
                               "Funding person or institution of the project",
                               "",
                               Datatype.PERSON_OR_ORGANIZATION,
                               Cardinality.ONE_TO_UNBOUND,
                               predicate=dsp_repo.hasFunder)

        self.grant = Property(meta, "Grant",
                              "Grant of the project",
                              "",
                              Datatype.GRANT,
                              predicate=dsp_repo.hasGrant)

        self.url = Property(meta, "URL",
                            "Landing page or Website of the project. We recommend DSP Landing Page.\nOptionally, a second URL can be added too.",
                            "https://test.dasch.swiss/",
                            Datatype.URL,
                            Cardinality.ONE_TO_TWO,
                            predicate=dsp_repo.hasURL)

        self.shortcode = Property(meta, "Shortcode",
                                  "Internal shortcode of the project",
                                  "0000",
                                  Datatype.SHORTCODE,
                                  Cardinality.ONE,
                                  value=shortcode,
                                  predicate=dsp_repo.hasShortcode)

        self.alternateName = Property(meta, "Alternate Name",
                                      "Alternative name of the project, e.g. in case of an overly long official name",
                                      "Another Title",
                                      Datatype.STRING,
                                      predicate=dsp_repo.hasAlternateName)

        self.dataManagementPlan = Property(meta, "Data Management Plan",
                                           "Data Management Plan of the project",
                                           "",
                                           Datatype.DATA_MANAGEMENT_PLAN,
                                           Cardinality.ZERO_OR_ONE,
                                           predicate=dsp_repo.hasDataManagementPlan)

        self.publication = Property(meta, "Publications",
                                    "Publications produced during the lifetime of the project",
                                    "Doe, J. (2000). A Publication.",
                                    Datatype.STRING,
                                    predicate=dsp_repo.hasPublication,
                                    multiline=True)

        self.contactPoint = Property(meta, "Contact Point",
                                     "Contact information",
                                     "",
                                     Datatype.PERSON_OR_ORGANIZATION,
                                     Cardinality.ZERO_OR_ONE,
                                     predicate=dsp_repo.hasContactPoint)

    def get_properties(self):
        return [
            self.name,
            self.shortcode,
            self.url,
            self.description,
            self.keywords,
            self.discipline,
            self.startDate,
            self.endDate,
            self.temporalCoverage,
            self.spatialCoverage,
            self.funder,
            self.grant,
            self.alternateName,
            self.dataManagementPlan,
            self.publication,
            self.contactPoint
        ]

    def __str__(self):
        return str(self.get_rdf_iri())


class Dataset(DataClass):
    """
    Dataset Shape.

    Corresponds to `dsp-repo:Dataset` in the ontology.
    """

    def __init__(self, name, project, meta):
        self.meta = meta
        self.title = Property(meta, "Title",
                              "Title of the dataset",
                              "Dataset-Title",
                              Datatype.STRING,
                              Cardinality.ONE,
                              value=f"Dataset of {name}",
                              predicate=dsp_repo.hasTitle)

        self.alternativeTitle = Property(meta, "Alternative Title",
                                         "Alternative title of the dataset",
                                         "Another Dataset-Title",
                                         Datatype.STRING,
                                         Cardinality.ZERO_OR_ONE,
                                         predicate=dsp_repo.hasAlternativeTitle)

        self.abstract = Property(meta, "Abstract",
                                 "Description of the dataset",
                                 "This is merely an exemplary dataset",
                                 Datatype.STRING_OR_URL,
                                 Cardinality.ONE_TO_UNBOUND,
                                 predicate=dsp_repo.hasAbstract,
                                 multiline=True)

        self.sameAs = Property(meta, "Alternative URL",
                               "Alternative URL to the dataset, if applicable",
                               "https://test.dasch.swiss/",
                               Datatype.URL,
                               Cardinality.UNBOUND,
                               predicate=dsp_repo.sameAs)

        self.typeOfData = Property(meta, "Type of Data",
                                   "Type of data related to the dataset",
                                   "xml",
                                   Datatype.CONTROLLED_VOCABULARY,
                                   Cardinality.ONE_TO_UNBOUND,
                                   value_options=["XML", "Text",
                                                  "Image", "Movie", "Audio"],
                                   predicate=dsp_repo.hasTypeOfData)

        self.documentation = Property(meta, "Documentation",
                                      "Additional documentation",
                                      '"http://www.example.org/documentation.md" or "Work in Progress"',
                                      Datatype.STRING_OR_URL,
                                      Cardinality.UNBOUND,
                                      predicate=dsp_repo.hasDocumentation)

        self.license = Property(meta, "License",
                                "The license terms of the dataset",
                                "https://creativecommons.org/licenses/by/3.0",
                                Datatype.URL,
                                Cardinality.ONE_TO_UNBOUND,
                                predicate=dsp_repo.hasLicense)

        self.accessConditions = Property(meta, "Conditions of Access",
                                         "Access conditions of the data",
                                         "Open Access",
                                         Datatype.STRING,
                                         Cardinality.ONE,
                                         predicate=dsp_repo.hasConditionsOfAccess)

        self.howToCite = Property(meta, "How to Cite",
                                  "How to cite the data",
                                  "Test-project (test), 2002, https://test.dasch.swiss",
                                  Datatype.STRING,
                                  Cardinality.ONE,
                                  predicate=dsp_repo.hasHowToCite)

        self.status = Property(meta, "Dataset Status",
                               "Current status of a dataset",
                               "The dataset is work in progress",
                               Datatype.CONTROLLED_VOCABULARY,
                               Cardinality.ONE,
                               value_options=['In planning', 'Ongoing', 'On hold', 'Finished'],
                               predicate=dsp_repo.hasStatus)

        self.datePublished = Property(meta, "Date Published",
                                      "Date of publication",
                                      "2000-08-01",
                                      Datatype.DATE,
                                      Cardinality.ZERO_OR_ONE,
                                      predicate=dsp_repo.hasDatePublished)

        self.language = Property(meta, "Language",
                                 "Language(s) of the dataset",
                                 "English",
                                 Datatype.STRING,
                                 Cardinality.ONE_TO_UNBOUND,
                                 predicate=dsp_repo.hasLanguage)

        self.project = Property(meta, "is Part of",
                                "The project to which the data set belongs",
                                "",
                                Datatype.PROJECT,
                                Cardinality.ONE,
                                value=project,
                                predicate=dsp_repo.isPartOf)

        self.attribution = Property(meta, "Qualified Attribution",
                                    "Persons/Organization involved in the creation of the dataset",
                                    '<person> + "editor"',
                                    Datatype.ATTRIBUTION,
                                    Cardinality.ONE_TO_UNBOUND,
                                    predicate=dsp_repo.hasQualifiedAttribution)

        self.dateCreated = Property(meta, "Date Created",
                                    "Creation of the dataset",
                                    "2000-08-01",
                                    Datatype.DATE,
                                    Cardinality.ZERO_OR_ONE,
                                    predicate=dsp_repo.hasDateCreated)

        self.dateModified = Property(meta, "Date Modified",
                                     "Last modification of the dataset",
                                     "2000-08-01",
                                     Datatype.DATE,
                                     Cardinality.ZERO_OR_ONE,
                                     predicate=dsp_repo.hasDateModified)

        self.distribution = Property(meta, "Distribution",
                                     "A downloadable form of this dataset, at a specific location, in a specific format",
                                     "https://test.dasch.swiss",
                                     Datatype.DOWNLOAD,
                                     Cardinality.ZERO_OR_ONE,
                                     predicate=dsp_repo.hasDistribution)

    def get_properties(self):
        return [
            self.project,
            self.title,
            self.abstract,
            self.language,
            self.typeOfData,
            self.attribution,
            self.license,
            self.howToCite,
            self.accessConditions,
            self.status,
            self.sameAs,
            self.alternativeTitle,
            self.documentation,
            self.datePublished,
            self.dateCreated,
            self.dateModified,
            self.distribution,
        ]

    def __str__(self):
        classname = str(self.get_rdf_iri()).split('#')[1]
        n1 = "<title missing>"
        if self.title.value:
            n1 = self.title.value
        return f"{classname}: {n1}"


class Person(DataClass):
    """
    Person Shape.

    Corresponds to `dsp-repo:Person` in the ontology.
    """

    def __init__(self, meta):
        self.meta = meta
        self.sameAs = Property(meta, "Alternative URL",
                               "Alternative URL, pointing to an authority file (ORCID, VIAF, GND, ...)",
                               "https://orcid.org/000-000-000-000",
                               Datatype.URL,
                               Cardinality.UNBOUND,
                               predicate=dsp_repo.sameAs)

        self.givenName = Property(meta, "Given Name",
                                  "Given name of the person",
                                  "John",
                                  Datatype.STRING,
                                  Cardinality.ONE_TO_UNBOUND_ORDERED,
                                  predicate=dsp_repo.hasGivenName)

        self.familyName = Property(meta, "Family Name",
                                   "Family name of the person. (Note that you can separate multiple family names with ';' if need be)",
                                   "Doe",
                                   Datatype.STRING,
                                   Cardinality.ONE,
                                   predicate=dsp_repo.hasFamilyName)

        self.email = Property(meta, "E-mail",
                              "E-mail address of the person",
                              "john.doe@dasch.swiss",
                              Datatype.EMAIL,
                              Cardinality.ZERO_TO_TWO,
                              predicate=dsp_repo.hasEmail)

        self.address = Property(meta, "Address",
                                "Postal address of the person",
                                "",
                                Datatype.ADDRESS,
                                Cardinality.UNBOUND,
                                predicate=dsp_repo.hasAddress)

        self.memberOf = Property(meta, "Member of",
                                 "Affiliation of the person",
                                 "",
                                 Datatype.ORGANIZATION,
                                 Cardinality.ONE_TO_UNBOUND,
                                 predicate=dsp_repo.isMemberOf)

        self.jobTitle = Property(meta, "Job Title",
                                 "Position/Job title of the person",
                                 "Dr.",
                                 Datatype.STRING,
                                 Cardinality.ONE_TO_UNBOUND,
                                 predicate=dsp_repo.hasJobTitle)

    def get_properties(self):
        return [
            self.familyName,
            self.givenName,
            self.memberOf,
            self.jobTitle,
            self.email,
            self.sameAs,
            self.address,
        ]

    def __str__(self):
        classname = str(self.get_rdf_iri()).split('#')[1]
        # classname = str(self.get_rdf_iri())
        n1 = "<first name missing>"
        if self.givenName.value:
            n1 = " ".join(self.givenName.value)
        n2 = "<family name missing>"
        if self.familyName.value:
            n2 = self.familyName.value
        return f"{classname}: {n1} {n2}"


class Organization(DataClass):
    """
    Organization Shape.

    Corresponds to `dsp-repo:Organization` in the ontology.
    """

    def __init__(self, meta):
        self.meta = meta

        self.name = Property(meta, "Legal Name",
                             "Legal name of the organization",
                             "DaSCH",
                             Datatype.STRING,
                             Cardinality.ONE_TO_UNBOUND,
                             predicate=dsp_repo.hasName)

        self.email = Property(meta, "E-mail",
                              "E-mail address of the organization",
                              "info@dasch.swiss",
                              Datatype.EMAIL,
                              Cardinality.ZERO_OR_ONE,
                              predicate=dsp_repo.hasEmail)

        self.address = Property(meta, "Address",
                                "Postal address of the organization",
                                "",
                                Datatype.ADDRESS,
                                Cardinality.UNBOUND,
                                predicate=dsp_repo.hasAddress)

        self.url = Property(meta, "URL",
                            "URL of the organization",
                            "https://dasch.swiss",
                            Datatype.URL,
                            Cardinality.ZERO_OR_ONE,
                            predicate=dsp_repo.hasURL)

    def get_properties(self):
        return [
            self.name,
            self.email,
            self.url,
            self.address,
        ]

    def __str__(self):
        classname = str(self.get_rdf_iri()).split('#')[1]
        n1 = "<name missing>"
        if self.name.value:
            n1 = " / ".join(self.name.value)
        return f"{classname}: {n1}"


class Grant(DataClass):
    """
    Grant Shape.

    Corresponds to `dsp-repo:Grant` in the ontology.
    """

    def __init__(self, meta):
        self.meta = meta

        self.name = Property(meta, "Name",
                             "Name of the grant",
                             "Ambizione",
                             Datatype.STRING,
                             Cardinality.ZERO_OR_ONE,
                             predicate=dsp_repo.hasName)

        self.url = Property(meta, "URL",
                            "URL of the grant",
                            "https://www.snf.ch/grants/001",
                            Datatype.URL,
                            Cardinality.ZERO_OR_ONE,
                            predicate=dsp_repo.hasURL)

        self.number = Property(meta, "Number",
                               "The number of the grant.",
                               "00012345",
                               Datatype.STRING,
                               Cardinality.ZERO_OR_ONE,
                               predicate=dsp_repo.hasNumber)

        self.funder = Property(meta, "Funder",
                               "Funding person or institution of the project",
                               "",
                               Datatype.PERSON_OR_ORGANIZATION,
                               Cardinality.ONE_TO_UNBOUND,
                               predicate=dsp_repo.hasFunder)

    def get_properties(self):
        return [
            self.funder,
            self.name,
            self.number,
            self.url,
        ]

    def __str__(self):
        classname = str(self.get_rdf_iri()).split('#')[1]
        n1 = "<funder missing>"
        v = self.funder.value
        if v:
            if isinstance(v, list):
                v = v[0]
            n1 = str(v)
        return f"{classname}: [{n1}]"


class Property():
    """
    General representation of a property.

    Corresponds to `sh:property`
    """

    def __init__(self, meta: MetaDataSet, name: str, description: str, example: str, datatype: Datatype.STRING,
                 cardinality=Cardinality.UNBOUND, value=None, value_options=None,
                 predicate=dsp_repo.whatever, multiline=False):
        self.meta = meta
        self.name = name
        self.description = description
        self.example = example
        self.datatype = datatype
        self.cardinality = cardinality
        self.value = value
        self.value_options = value_options
        self.predicate = predicate
        self.multiline = multiline

    def get_url_property_id(url: str) -> str:
        """
        This method tries to guess the propetyID for a URL.

        For certain pre-defined cases, a reasonable propertyID is chosen;
        otherwise, the net location is being extracted, if possible.

        Args:
            url (str): a URL

        Returns:
            str: a propertyID
        """
        # LATER: remove this method with next breaking change, and leave it to the utils entirely
        return utils.get_url_property_id(url)

    def get_triples(self, subject: URIRef) -> Graph:
        """
        Returns a Graph containing the triples that represent this property, in respect to a given subject.

        Args:
            subject (URIRef): the subject to which the property is object

        Returns:
            Graph: a graph containing one or multiple triples that represent the property
        """
        g = Graph()
        # Ensure the data can be looped
        vals = self.value
        if not isinstance(vals, list):
            vals = [vals]
        if self.datatype == Datatype.STRING and \
                self.cardinality == Cardinality.ONE_TO_UNBOUND_ORDERED and \
                vals and vals[0]:
            vals = [';'.join(vals)]
        for v in vals:
            if not v:
                continue
            if isinstance(v, str) and v.isspace():
                continue
            # resolve datatype ambiguity
            datatype = self.datatype
            if datatype == Datatype.STRING_OR_URL:
                if v and validators.url(str(v)):
                    datatype = Datatype.URL
                elif v and v.startswith('www.'):
                    v = "http://" + v
                    datatype = Datatype.URL
                else:
                    datatype = Datatype.STRING
            if datatype == Datatype.PERSON_OR_ORGANIZATION:
                if isinstance(v, Person):
                    datatype = Datatype.PERSON
                else:
                    datatype = Datatype.ORGANIZATION
            # Handle datatypes
            if datatype == Datatype.STRING \
                    or datatype == Datatype.CONTROLLED_VOCABULARY \
                    or datatype == Datatype.SHORTCODE:
                g.add((subject, self.predicate, Literal(v, datatype=XSD.string)))
            elif datatype == Datatype.DATE:
                g.add((subject, self.predicate, Literal(v, datatype=XSD.date)))
            elif datatype == Datatype.URL:
                blank = BNode()
                g.add((subject, self.predicate, blank))
                g.add((blank, RDF.type, SDO.URL))
                b2 = BNode()
                g.add((blank, SDO.propertyID, b2))
                g.add((b2, RDF.type, SDO.PropertyValue))
                g.add((b2, SDO.propertyID, Literal(utils.get_url_property_id(url=v))))
                g.add((blank, SDO.url, Literal(v)))
            elif datatype == Datatype.PLACE:
                b0 = BNode()
                g.add((subject, self.predicate, b0))
                g.add((b0, RDF.type, SDO.Place))
                b1 = BNode()
                g.add((b0, SDO.url, b1))
                g.add((b1, RDF.type, SDO.URL))
                g.add((b1, SDO.url, Literal(v)))
                b2 = BNode()
                g.add((b1, SDO.propertyID, b2))
                g.add((b2, RDF.type, SDO.PropertyValue))
                g.add((b2, SDO.propertyID, Literal(utils.get_url_property_id(v), datatype=XSD.string)))
            elif datatype == Datatype.PERSON:
                g.add((subject, self.predicate, v.get_rdf_iri()))
            elif datatype == Datatype.ORGANIZATION:
                g.add((subject, self.predicate, v.get_rdf_iri()))
            elif datatype == Datatype.PROJECT:
                g.add((subject, self.predicate, v.get_rdf_iri()))
            elif datatype == Datatype.DATA_MANAGEMENT_PLAN:
                if v[0] or v[1]:
                    try:
                        dmp = URIRef(dsp_repo[f'dsp-{self.meta.shortcode}-dmp'])
                    except Exception as e:
                        print(f'Warning: DMP has non-unique IRI ({e})')
                        # LATER: this should not be necessary anymore. remove with next breaking changes
                        dmp = URIRef(dsp_repo['dmp'])
                    g.add((subject, self.predicate, dmp))
                    g.add((dmp, RDF.type, dsp_repo.DataManagementPlan))
                    if v[0]:
                        g.add((dmp, dsp_repo.isAvailable, Literal(
                            v[0], datatype=XSD.boolean)))
                    if v[1]:
                        b1 = BNode()
                        g.add((dmp, dsp_repo.hasURL, b1))
                        g.add((b1, RDF.type, SDO.URL))
                        g.add((b1, SDO.url, Literal(v[1])))
            elif datatype == Datatype.ADDRESS:
                if not v[0] and not v[1] and not v[2]:
                    return g
                b0 = BNode()
                g.add((subject, self.predicate, b0))
                g.add((b0, RDF.type, SDO.PostalAddress))
                g.add((b0, SDO.streetAddress, Literal(
                    v[0], datatype=XSD.string)))
                g.add((b0, SDO.postalCode, Literal(v[1], datatype=XSD.string)))
                g.add((b0, SDO.addressLocality, Literal(
                    v[2], datatype=XSD.string)))
            elif datatype == Datatype.GRANT:
                g.add((subject, self.predicate, v.get_rdf_iri()))
            elif datatype == Datatype.ATTRIBUTION:
                b0 = BNode()
                g.add((subject, self.predicate, b0))
                g.add((b0, RDF.type, prov.Attribution))
                g.add((b0, dsp_repo.hasRole, Literal(v[0], datatype=XSD.string)))
                g.add((b0, prov.agent, v[1].get_rdf_iri()))
            elif datatype == Datatype.DOWNLOAD:
                b0 = BNode()
                g.add((subject, self.predicate, b0))
                g.add((b0, RDF.type, SDO.DataDownload))
                g.add((b0, SDO.url, Literal(v)))
            elif datatype == Datatype.EMAIL:
                if isinstance(v, tuple):
                    if v and v[0]:
                        g.add((subject, self.predicate, Literal(v[0])))
                    if v and v[1]:
                        g.add((subject, self.predicate, Literal(v[1])))
                else:
                    if v and not v.isspace():
                        g.add((subject, self.predicate, Literal(v)))
            else:
                print(f"{datatype}: {v}\n-> don't know how to serialize this.\n")
        return g

    def validate(self) -> (Validity, str):
        datatype = self.datatype
        cardinality = self.cardinality
        value = self.value

        missing = "Required value is missing."
        valid = "The current value is valid."
        optional = "This field is optional"
        no_url = "Invalid URL"
        no_mail = "Invalid e-mail address"

        if not value:
            if Cardinality.isMandatory(cardinality):
                return Validity.REQUIRED_VALUE_MISSING, missing
            else:
                return Validity.OPTIONAL_VALUE_MISSING, valid

        if datatype == Datatype.STRING or \
                datatype == Datatype.STRING_OR_URL or \
                datatype == Datatype.DOWNLOAD:
            if cardinality == Cardinality.ONE:
                if value and not value.isspace():
                    return Validity.VALID, valid
                else:
                    return Validity.REQUIRED_VALUE_MISSING, missing
            elif cardinality == Cardinality.ONE_TO_TWO:
                if value[0] and not value[0].isspace():
                    return Validity.VALID, valid
                else:
                    return Validity.REQUIRED_VALUE_MISSING, missing
            elif cardinality == Cardinality.ZERO_OR_ONE:
                if value and not value.isspace():
                    return Validity.VALID, valid
                else:
                    return Validity.OPTIONAL_VALUE_MISSING, optional
            elif cardinality == Cardinality.ONE_TO_UNBOUND or \
                    cardinality == Cardinality.ONE_TO_UNBOUND_ORDERED:
                if len(value) > 0 and value[0] and not value[0].isspace():
                    return Validity.VALID, valid
                else:
                    return Validity.REQUIRED_VALUE_MISSING, missing
            elif cardinality == Cardinality.UNBOUND:
                if len(value) > 0 and value[0] and not value[0].isspace():
                    return Validity.VALID, valid
                else:
                    return Validity.OPTIONAL_VALUE_MISSING, optional

        elif datatype == Datatype.SHORTCODE:
            if re.match('[a-zA-Z0-9]{4}$', value):
                return Validity.VALID, valid
            else:
                return Validity.INVALID_VALUE, "Shortcode must be exactly 4 alphanumeric characters."

        elif datatype == Datatype.URL or \
                datatype == Datatype.PLACE:
            if cardinality == Cardinality.UNBOUND:
                if len(value) > 0 and value[0] and not value[0].isspace():
                    if utils.areURLs(value):
                        return Validity.VALID, valid
                    else:
                        return Validity.INVALID_VALUE, no_url
                else:
                    return Validity.OPTIONAL_VALUE_MISSING, optional
            elif cardinality == Cardinality.ZERO_OR_ONE:
                if value and not value.isspace():
                    if utils.isURL(value):
                        return Validity.VALID, valid
                    else:
                        return Validity.INVALID_VALUE, no_url
                else:
                    return Validity.OPTIONAL_VALUE_MISSING, optional
            elif cardinality == Cardinality.ONE_TO_TWO or \
                    cardinality == Cardinality.ONE_TO_UNBOUND:
                if value[0] and not value[0].isspace():
                    if utils.areURLs(value):
                        return Validity.VALID, valid
                    else:
                        return Validity.INVALID_VALUE, no_url
                else:
                    return Validity.REQUIRED_VALUE_MISSING, missing

        elif datatype == Datatype.EMAIL:
            if cardinality == Cardinality.ZERO_OR_ONE:
                if value and not value.isspace():
                    if utils.is_email(value):
                        return Validity.VALID, valid
                    else:
                        return Validity.INVALID_VALUE, no_mail
                else:
                    return Validity.OPTIONAL_VALUE_MISSING, optional
            elif cardinality == Cardinality.ZERO_TO_TWO:
                if value[0] and not value[0].isspace():
                    if utils.are_emails(value):
                        return Validity.VALID, valid
                    else:
                        return Validity.INVALID_VALUE, no_mail
                else:
                    return Validity.OPTIONAL_VALUE_MISSING, optional

        elif datatype == Datatype.GRANT or \
                datatype == Datatype.PROJECT or \
                datatype == Datatype.PERSON or \
                datatype == Datatype.ORGANIZATION or \
                datatype == Datatype.PERSON_OR_ORGANIZATION:
            if cardinality == Cardinality.UNBOUND:
                if len(value) > 0 and value[0]:
                    return Validity.VALID, valid
                else:
                    return Validity.OPTIONAL_VALUE_MISSING, optional
            if cardinality == Cardinality.ONE_TO_UNBOUND:
                if len(value) > 0 and value[0]:
                    return Validity.VALID, valid
                else:
                    return Validity.REQUIRED_VALUE_MISSING, missing
            if cardinality == Cardinality.ONE:
                if value:
                    return Validity.VALID, valid
                else:
                    return Validity.REQUIRED_VALUE_MISSING, missing
            if cardinality == Cardinality.ZERO_OR_ONE:
                if value:
                    return Validity.VALID, valid
                else:
                    return Validity.OPTIONAL_VALUE_MISSING, optional

        elif datatype == Datatype.DATE:
            if not value or value.isspace():
                if cardinality == Cardinality.ONE:
                    return Validity.REQUIRED_VALUE_MISSING, missing
                elif cardinality == Cardinality.ZERO_OR_ONE:
                    return Validity.OPTIONAL_VALUE_MISSING, optional
            elif value and re.match(r'\d{4}-\d{2}-\d{2}$', value):
                return Validity.VALID, valid
            else:
                return Validity.INVALID_VALUE, "Not a valid date."

        elif datatype == Datatype.ADDRESS:
            if cardinality == Cardinality.UNBOUND:
                if value[0] and not value[0].isspace() and \
                        value[1] and not value[1].isspace() and \
                        value[2] and not value[2].isspace():
                    return Validity.VALID, valid
                elif (not value[0] or value[0].isspace()) and \
                        (not value[1] or value[1].isspace()) and \
                        (not value[2] or value[2].isspace()):
                    return Validity.OPTIONAL_VALUE_MISSING, optional
                else:
                    return Validity.INVALID_VALUE, "Not a valid address."

        elif datatype == Datatype.CONTROLLED_VOCABULARY:
            if cardinality == Cardinality.ONE_TO_UNBOUND:
                for v in value:
                    if v not in self.value_options:
                        return Validity.INVALID_VALUE, f"Value '{v}' not allowed."
                return Validity.VALID, valid
            elif cardinality == Cardinality.ONE:
                if value not in self.value_options:
                    return Validity.INVALID_VALUE, f"Value '{value}' not allowed."
                return Validity.VALID, valid

        elif datatype == Datatype.ATTRIBUTION:
            if cardinality == Cardinality.ONE_TO_UNBOUND:
                for v in value:
                    if (not v[0] or v[0].isspace()) or \
                            (not v[1] or not isinstance(v[1], (Person, Organization))):
                        return Validity.INVALID_VALUE, "Not a valid address."
                return Validity.VALID, valid

        elif datatype == Datatype.DATA_MANAGEMENT_PLAN:
            if cardinality == Cardinality.ZERO_OR_ONE:
                if value[0] or (value[1] and not value[1].isspace()):
                    return Validity.VALID, valid
                else:
                    return Validity.OPTIONAL_VALUE_MISSING, optional

        print(f'behavior undefined:\ncard: {cardinality}\ntype: {datatype}\n')
        return "", ""

    def __str__(self):
        if self.value:
            return str(self.value)
        return f"<Property '{self.name}' undefined>"
